import copy
import numpy as np
import opt_einsum
from acqdp.tensor_network import ContractionTask, defaultOrderResolver, ContractionCost


class Compiler:
    """
    Compile a :class:`ContractionScheme` indicating contraction scheme and sliced edges into a :class:`ContractionTask`, a hardware-aware format that can be readily used for tensor network contraction.

    Typically, a :class:`ContractionScheme` found by one of the :class:`OrderFinder` only aims to minimize the theoretical floating number operations and / or largest size of intermediate tensors.
    Although those are typically accurate indicators of the overall contracion cost, they fail to accomodate inefficiencies from elsewhere.
    The compiler takes care of machine-related inefficiencies by slightly modifying the order.

    :ivar do_patch: If set to false, the patching and reorganization of the contraction order below will be skipped.
    :ivar patch_size: Whenever two adjacent branches both have size less than the patch size, the branches are merged together. This is to avoid inefficiency called by skewed tensor shapes in tensor multiplication, with a slight sacrifice in floating number operations.
    :ivar reorg_thres: Threshold for contraction order reorganization. The order is seperated into small tensor multiplications and large tensor multiplications. All pairwise tensor multiplications involving tensors with size less than `reorg_thres` will be put forward whenever possible.
    """
    def __init__(self,
                 reorg_thres=23,
                 patch_size=5,
                 do_patch=False,
                 **kwargs):
        self.reorg_thres = reorg_thres
        self.patch_size = patch_size
        self.do_patch = do_patch

    def compile(self, tn, scheme, **kwargs):
        """
        :param tn: The :class:`TensorNetwork` the input contraction scheme is based on.
        :type tn: :class:`TensorNetwork`
        :param scheme: The contraction scheme from which the task is generated.
        :type scheme: :class:`ContractionScheme`

        :returns: :class:`ContractionTask` -- The compiled contraction task ready for execution.
        """
        tn = tn._expand_and_delta()
        res = self._generate_template(tn, scheme=scheme, **kwargs)
        res.set_data({
            node: tn.network.nodes[(0, node)]['tensor'].contract()
            for node in tn.nodes_by_name
        })
        res.update_fix(tn)
        return res

    def _generate_template(self, tn, scheme, test=0, **kwargs):
        inputs = {node[1]: [None] for node in tn.nodes}
        data_ref = {(0, node): inputs[node] for node in inputs}
        order, split = scheme.order, scheme.slices
        split = [(1, i) for i in split]
        tn_copy = tn.copy().expand(recursive=True)
        tn_copy.open_edges += [i[1] for i in split]
        tn_copy.fix()
        for edge_name in list(tn_copy.edges_by_name):
            if (len(tn_copy.network[(1, edge_name)])
                    == 0) and edge_name not in tn_copy.open_edges:
                tn_copy.network.remove_node((1, edge_name))
        init_order, final_order = self._order_patch(tn_copy.copy(), order, split,
                                                   **kwargs)
        k = self._generate_template_fix(tn, [], data_ref)
        if isinstance(k, ContractionTask):
            return k
        lst, fix_dict = k

        for item in init_order:
            if item[1] == '#':
                break
            stn_name, stn = tn_copy.encapsulate(item[0], stn_name=item[1])
            lst += self._generate_template_basic(stn, data_ref, (0, stn_name))

        cnt = len(lst)

        fix_lst, fix_dict_new = self._generate_template_fix(
            tn_copy, split, data_ref)
        lst += fix_lst
        fix_dict.update(fix_dict_new)

        tn_copy.open_edges = [
            i for i in tn_copy.open_edges if (1, i) not in split
            ]
        for i in split:
            tn_copy.fix_edge(i[1])
        tn_copy.fix()

        lst += self._generate_template_basic(tn_copy, data_ref, '#')
        _, _, shapes = tn_copy.subscripts()
        if len(final_order) > 0:
            path = []
            lst_nodes = list(n[1] for n in tn_copy.nodes)
            for o in final_order:
                path.append(
                    (lst_nodes.index(o[0][0]), lst_nodes.index(o[0][1])))
                lst_nodes.remove(o[0][0])
                lst_nodes.remove(o[0][1])
                lst_nodes.append(o[1])
            expr = opt_einsum.contract_expression(lst[-1][3]['subscripts'],
                                                  *shapes,
                                                  optimize=path)
            lst[-1][3]['expr'] = expr
        res = ContractionTask(commands=lst,
                               fix_dict=fix_dict,
                               inputs=inputs,
                               output=data_ref['#'],
                               shape=tn.shape,
                               open_edges=tn.open_edges,
                               sub_outputs=tn_copy.open_edges,
                               cnt=cnt)
        return res

    def _generate_template_fix(self, tn, split, data_ref):
        lst = []

        fix_dict = self._generate_template_edge_fix(tn, split)
        if not isinstance(fix_dict, dict):
            return fix_dict
        for node in tn.nodes:
            fix_idx = []
            fix_to = []
            for i, edge in enumerate(tn.network.nodes[node]['edges']):
                if (1, edge) in fix_dict:
                    fix_idx.append(i)
                    fix_to.append(fix_dict[(1, edge)][1])
            if len(fix_idx) > 0:
                new_res = [None]
                lst.append(('f', data_ref[node], new_res, {'fix_idx': fix_idx, 'fix_to': fix_to}))
                data_ref[node] = new_res
        res_dict = {(k, tuple(fix_dict[k][0])): fix_dict[k][1] for k in fix_dict}
        return lst, res_dict

    def _generate_template_edge_fix(self, tn, split):
        fix_dict = {}
        for edge in tn.edges:
            if edge in split:
                tn.network.nodes[edge]['fix_to'] = list(range(tn.network.nodes[edge]['dim']))
                fix_dict[edge] = (list(range(tn.network.nodes[edge]['dim'])), [0])
            elif 'fix_to' in tn.network.nodes[edge]:
                ll = len(tn.network.nodes[edge]['fix_to'])
                if ll == 0:
                    return ContractionTask(output=[(0, np.zeros(tn.shape))])
                else:
                    fix_dict[edge] = (list(tn.network.nodes[edge]['fix_to']), list(tn.network.nodes[edge]['fix_to']))
        return fix_dict

    def _generate_template_delta(self, tn, data_ref):
        open_edge_names = {}
        edges = tn.edges_by_name_from_nodes_by_name()
        if len(tn.nodes) == 0:
            id_node = np.array(1.)
            tn.add_node('*', [], id_node)
            data_ref[(0,'*')] = [(0, id_node)]
        for (i, edge) in enumerate(tn.open_edges):
            if edge not in open_edge_names:
                open_edge_names.update({edge: 0})
            else:
                open_edge_names[edge] += 1
                new_node = np.eye(tn.network.nodes[(1, edge)]['dim'])
                data_ref[(0, ("*", edge, open_edge_names[edge]))] = [(0, new_node)]
                tn.add_node(("*", edge, open_edge_names[edge]), [edge, (edge, open_edge_names[edge])], new_node)
                tn.open_edges[i] = (edge, open_edge_names[edge])
        for edge in open_edge_names:
            diag = None
            if edge not in edges and open_edge_names[edge] == 0:
                diag = [1.] * tn.network.nodes[(1, edge)]['dim']
            if 'fix_to' in tn.network.nodes[(1, edge)]:
                diag = [1. if i in tn.network.nodes[(1, edge)]['fix_to'] else 0 for i in range(tn.network.nodes[(1, edge)]['dim'])]
                tn.network.nodes[(1, edge)].pop('fix_to')
            if diag is not None:
                new_node = np.array(diag)
                data_ref[(0, ('*', edge, 0))] = [(0, new_node)]
                tn.add_node(('*', edge, 0), [edge], new_node)

    def _generate_template_basic(self, tn, data_ref, new_name, **kwargs):
        new_res = [None]
        data_ref[new_name] = new_res
        if new_name == '#':
            self._generate_template_delta(tn, data_ref)
        lhs = [data_ref[node] for node in tn.nodes]

        return [('c', lhs, new_res, self._generate_template_einsum(tn, subscripts=tn.subscripts(), **kwargs), [node for node in tn.nodes])]

    def _generate_template_einsum(self, tn, subscripts, **kwargs):
        lhs_str, rhs_str, _ = subscripts
        command = ','.join(lhs_str) + '->' + rhs_str
        return {'subscripts': command,  'dtype': tn.dtype}

    def _order_patch(self, tn, order, split_edges, **kwargs):
        if self.do_patch:
            from acqdp.tensor_network.contraction_tree import ContractionTree
            tn_copy = tn.copy()
            tn_copy.open_edges += [i[1] for i in split_edges]
            tn_copy.fix()
            for edge_name in list(tn_copy.edges_by_name):
                if (len(tn_copy.network[(1, edge_name)])
                        == 0) and edge_name not in tn_copy.open_edges:
                    tn_copy.network.remove_node((1, edge_name))
            assert tn_copy.dtype == tn.dtype
            order = ContractionTree.from_order(tn_copy, order).full_order
            init_order, final_order = self._reorg_order(tn_copy, order)
            tn_copy.open_edges = [
                i for i in tn_copy.open_edges if (1, i) not in split_edges
            ]
            for i in split_edges:
                tn_copy.fix_edge(i[1])
            tn_copy.fix()
            if len(final_order) > 0:
                final_order = self._patch_order(tn_copy, final_order)
            return init_order, final_order
        else:
            return [], order

    def _reorg_order(self, graph, order):
        graph.fix()
        size_dic = {}
        subs, order = defaultOrderResolver.order_to_subscripts_tree(
            graph, order)
        assert len(subs) == len(order), f'{len(subs)}, {len(order)}'
        for i, o in enumerate(order):
            size_dic[o[1]] = len(subs[i][1])
            for j, k in enumerate(o[0]):
                size_dic[k] = len(subs[i][0][j])
        init_order = []
        node_set = set([a[1] for a in graph.nodes])
        while True:
            for o in order:
                if size_dic[o[1]] < self.reorg_thres and np.all(
                    [a in node_set for a in o[0]]):
                    init_order.append(o)
                    order.remove(o)
                    for a in o[0]:
                        node_set.remove(a)
                    node_set.add(o[1])
                    graph.encapsulate(nodes=o[0], stn_name=o[1])
                    break
            else:
                break
        return init_order, order

    def _patch_order(self, graph, order):
        from acqdp.tensor_network.contraction_tree import ContractionTree
        graph_copy = graph.copy()
        graph_copy.fix()
        tree = ContractionTree.from_order(graph_copy, order)

        res = tree.patch(self.patch_size)
        return res



from .circuit import (Operation,
                              ImmutableOperation,
                              PureOperation,
                              Circuit,
                              ControlledOperation,
                              Controlled,
                              ControlledCircuit,
                              CompState,
                              CompMeas,
                              SuperPosition)

from acqdp.tensor_network.tensor_valued import TensorValued

from acqdp.tensor_network.tensor_sum import TensorSum

from acqdp.tensor_network.tensor_network import TensorNetwork

import numpy as np


class Converter:
    """
    Generic converter class, with method mapping a circuit to
    a tensor network.
    """
    @classmethod
    def convert_pure(cls, operation: Operation) -> TensorValued:
        """
        Do a pure conversion for noiseless circuit described as a :class:`Operation` object to a tensor network as a :class:`TensorValued` object.

        :param operation: the noisy circuit to convert.
        :type operation: :class:`Operation`.
        :returns:  :class:`TensorValued` -- the tensor network to describe the input noisy circuit.

        """
        if not operation.is_pure:
            raise ValueError("Invalid operation: trying to do pure conversion for noisy circuit")
        if isinstance(operation, ControlledCircuit):
            op = operation.circuit
            c = Circuit()
            qubits = op.all_qubits
            for gate in op.operations_by_name:
                c.append(Controlled(op.operations_by_name[gate]["operation"],
                                    conditioned_on=operation.conditioned_on),
                         [0] + [1 + qubits.index(i) for i in op.operations_by_name[gate]['qubits']])
            return c.tensor_pure
        elif isinstance(operation, ControlledOperation):
            tn = TensorNetwork()
            tn.add_node("TC", tensor=operation._tensor_control)
            out_edges = []
            in_edges = []
            lst_ctrl, len_ctrl = operation._indices_with_property("^c$")
            lst_out, len_out = operation._indices_with_property(".*o$")
            lst_in, len_in = operation._indices_with_property("^i")
            for i in operation._output_indices[0]:
                if i in lst_ctrl:
                    out_edges.append(lst_ctrl.index(i))
                else:
                    out_edges.append(len_ctrl + lst_out.index(i))
            for i in operation._input_indices[0]:
                if i in lst_ctrl:
                    in_edges.append(lst_ctrl.index(i))
                else:
                    in_edges.append(len_ctrl + len_out + lst_in.index(i))
            tn.open_edges = out_edges + in_edges
            return tn
        elif isinstance(operation, PureOperation):
            return operation._tensor_pure
        elif isinstance(operation, SuperPosition):
            ts = TensorSum()
            for i in range(len(operation.coefs)):
                ts.add_term(i, operation.coefs[i] * operation.operations[i].tensor_pure)
            return ts
        else:
            if not isinstance(operation, Circuit):
                raise ValueError("Invalid operation: operation is not yet supported")
            # This cannot happen normally because Circuit.is_pure checks
            # Circuit.is_valid, but is left in just in case someone subclasses
            # Circuit and overrides is_pure.
            if not operation.is_valid:  # pragma: no cover
                raise ValueError("Invalid operation: trying to convert an invalid circuit into a tensor network")
            dic = {}
            dic_in = {}
            dic_out = {}
            tn = TensorNetwork()
            all_edges = set()
            for time_step in list(operation.operations_by_time)[::-1]:
                for idx, gate in enumerate(operation.operations_by_time[time_step]):
                    op = operation.operations_by_name[gate]['operation']
                    qubits = operation.operations_by_name[gate]['qubits']
                    out_edges = []
                    in_edges = []
                    for i in op._output_indices[0]:
                        q = qubits[i]
                        if q not in dic:
                            dic_out[q] = (time_step+1, idx, q)
                            out_edges.append(dic_out[q])
                        else:
                            edge = dic.pop(q)
                            out_edges.append(edge)
                            dic_in.pop(q)
                    for i in op._input_indices[0]:
                        q = qubits[i]
                        dic[q] = (time_step, idx, q)
                        in_edges.append(dic[q])
                        dic_in[q] = dic[q]
                    for edge in out_edges + in_edges:
                        if edge not in all_edges:
                            tn.add_edge(edge, bond_dim=2)
                            all_edges.add(edge)
                    if op in CompState:
                        tn.fix_edge(out_edges[0], CompState.index(op))
                    elif op in CompMeas:
                        tn.fix_edge(in_edges[0], CompMeas.index(op))
                    else:
                        tn.add_node(gate, out_edges + in_edges, op.tensor_pure)
            tn.open_edges = [dic_out[i] for i in sorted(dic_out)] +\
                            [dic_in[i] for i in sorted(dic_in)]
            return tn

    @classmethod
    def convert_control(cls, operation: PureOperation) -> TensorValued:
        """
        Do a control conversion for noisy circuit described as a :class:`Operation` object to a tensor network as a :class:`TensorValued` object.

        :param operation: the noisy circuit to convert.
        :type operation: :class:`Operation`.
        :returns:  :class:`TensorValued` -- the tensor network to describe the input noisy circuit.

        """
        if not operation.is_pure:
            raise ValueError("Noisy operation does not have pure block-diagonal form")
        if isinstance(operation, ControlledOperation):
            return operation._tensor_control
        else:
            return operation.tensor_pure

    @classmethod
    def convert_density(cls, operation: Operation) -> TensorValued:
        """
        Do a density conversion for noisy circuit described as a :class:`Operation` object to a tensor network as a :class:`TensorValued` object.

        :param operation: the noisy circuit to convert.
        :type operation: :class:`Operation`.
        :returns:  :class:`TensorValued` -- the tensor network to describe the input noisy circuit.

        """
        if isinstance(operation, PureOperation):
            tp = operation.tensor_pure
            tn = TensorNetwork()
            len_tensor = len(tp.shape)
            tn.add_node("TP", list(range(len_tensor)), tp)
            tn.add_node("~TP", list(range(len_tensor, 2 * len_tensor)), (~tp).expand())
            tn.open_edges = list(range(2 * len_tensor))
            return tn
        elif isinstance(operation, ImmutableOperation):
            return operation._tensor_density
        else:
            if not isinstance(operation, Circuit):
                raise ValueError("Invalid operation: operation is not yet supported")
            if not operation.is_valid:
                raise ValueError('Invalid operation')
            dic = {}
            tn = TensorNetwork()
            dic_out = {}
            dic_in = {}
            for time_step in list(operation.operations_by_time):
                for idx, gate in enumerate(operation.operations_by_time[time_step]):
                    op = operation.operations_by_name[gate]["operation"]
                    qubits = operation.operations_by_name[gate]['qubits']
                    out_edges_p = []
                    in_edges_p = []
                    out_edges_d = []
                    in_edges_d = []

                    for i in op._input_indices[0]:
                        q = qubits[i]
                        if q not in dic:
                            in_edges_p.append((time_step, idx, q, 'p'))
                            in_edges_d.append((time_step, idx, q, 'd'))
                            dic_in[q] = [time_step, idx, q]
                        else:
                            edge = list(dic.pop(q))
                            in_edges_p.append(tuple(edge + ['p']))
                            in_edges_d.append(tuple(edge + ['d']))
                            dic_out.pop(q)
                    for i in op._output_indices[0]:
                        q = qubits[i]
                        dic[q] = (time_step + 1, idx, q)
                        out_edges_p.append((time_step + 1, idx, q, 'p'))
                        out_edges_d.append((time_step + 1, idx, q, 'd'))
                        dic_out[q] = [time_step + 1, idx, q]
                    tn.add_node(gate, out_edges_p + in_edges_p + out_edges_d + in_edges_d, op.tensor_density)
            tn.open_edges = [tuple(dic_out[i] + ['p']) for i in sorted(dic_out)] +\
                            [tuple(dic_in[i] + ['p']) for i in sorted(dic_in)] +\
                            [tuple(dic_out[i] + ['d']) for i in sorted(dic_out)] +\
                            [tuple(dic_in[i] + ['d']) for i in sorted(dic_in)]
            return tn



import numpy as np
import collections
import re
from copy import copy
import math
# from .myabc import MN
from acqdp.tensor_network.tensor_valued import TensorValued
from acqdp.tensor_network.tensor import Tensor
from acqdp.tensor_network.tensor_network import TensorNetwork
from typing import List, Dict, Optional, Set


INDENT = "  "
"""The unit of indentation used for :meth:`Operation.tree_string`."""


class Operation(object):
    """Base class for quantum opertations.

    By itself, it can represent a generic quantum operation, although one would
    not be able to do much with it. Usually, one should use
    :class:`ImmutableOperation` (or more commonly, one of its subclasses) for
    simple quantum operations with an explicit tensor representation, and
    :class:`Circuit` for quantum operations better represented as a composition
    of simpler operations. (Technically an :class:`ImmutableOperation` can
    represent a complex operation by using a :class:`TensorNetwork` as the data,
    but that would be an uncommon use case.)

    :ivar name: Name of the quantum operation. Defaults to "GenOp".
    :vartype name: str, optional
    """

    def __init__(self, name="GenOp"):
        self.name = name

    def __str__(self) -> str:  # pragma: no cover
        return str(self.name)

    def tree_string(self, indent=0):  # pragma: no cover
        """Return an indented string that describes the operation.

        This is mainly used for visualization of :class:`Circuit` instances.
        Notably, the returned string should not include the name of the
        operation, which would already be included by default in the tree
        string of the "parent" operation.

        It is fine to return an empty string, but otherwise, the string should
        be indented with `indent` copies of the string :data:`INDENT`, and
        terminated with a newline.

        :param indent: The amount of indent needed. Defaults to 0.
        :type indent: int, optional
        """
        return ""

    def __repr__(self) -> str:  # pragma: no cover
        return repr(vars(self))

    def __mul__(self, other):
        """Return two operations in parallel as a quantum circuit.

        The resulting :class:`Circuit` will have a number of qubits :math:`n`
        equal to the total number of qubits in both operands, indexed by
        integers from 0 to :math:`n-1`. The qubits are ordered naturally, i.e.,
        qubits in the left operand come first, and within both operands the
        original qubit orders are preserved.

        Note that this function will regard both operands as atomic operations:
        No attempt is made to "expand" the operands even if one or both of them
        are themselves :class:`Circuit` instances.
        """
        if isinstance(other, Operation):
            return Circuit().append(self, list(range(len(self.shape))), 0).append(other, list(range(len(self.shape), len(self.shape) + len(other.shape))), 0)
        else:
            return NotImplemented

    def __or__(self, other):
        """Return the concatenation of two operations as a quantum circuit.

        The left operand will happen first. Note that this ordering convention
        is different from how the result would be represented as a product of
        matrices. For example, ``ZeroState | HGate | ZGate`` will result in the
        state :math:`ZH|0\\rangle`.

        If the left operand is a :class:`Circuit` instance, then the qubit names
        in it will be preserved, and the same will apply to the right operand if
        it is a :class:`Circuit` instance too; otherwise qubits in the right
        operand will be indexed by integers from 0. The qubit names will
        determine how the two circuits are connected.

        If the left operand is not a :class:`Circuit` instance, then the qubits
        in *both* operands will be indexed by integers from 0, and the circuits
        will be connected correspondingly.
        """
        if isinstance(self, Circuit):
            if isinstance(other, Circuit):
                copy_self = copy(self)
                for time_step in other.operations_by_time:
                    for op in other.operations_by_time[time_step]:
                        operation = other.operations_by_time[time_step][op]
                        copy_self.append(operation['operation'],
                                         operation['qubits'])
                return copy_self
            else:
                return copy(self).append(other, list(range(len(other.shape))))
        else:
            return Circuit().append(self, list(range(len(self.shape)))).append(other, list(range(len(other.shape))))

    def _indices_with_property(self, pattern):
        res = [i for i in range(len(self.shape))
               if bool(re.match(pattern, self.shape[i]))]
        return res, len(res)

    @property
    def _input_indices(self):
        return self._indices_with_property("^[ibc]")

    @property
    def _output_indices(self):
        return self._indices_with_property(".*[odc]$")

    @property
    def tensor_pure(self):
        """Convert the operation into a tensor network representing the action
        of the operation in the pure state picture.

        Examples of tensor representations in the pure state picture include
        state vectors of pure states, and unitary matrices of unitary gates.

        Raises an error if the operation is not pure.
        """
        from .converter import Converter
        return Converter.convert_pure(self)

    @property
    def tensor_density(self):
        """Convert the operation into a tensor network representing the action
        of the operation in the density matrix picture.

        Examples of tensor representations in the density matrix picture include
        density matrices of general quantum states, and Choi matrices of quantum
        operations. As a special case, when the operation is pure, the tensor
        network returned by :attr:`tensor_density` will consist of two disjoint
        components, one being the tensor network returned by :attr:`tensor_pure`
        and the other being its adjoint.
        """
        from .converter import Converter
        return Converter.convert_density(self)

    @property
    def tensor_control(self):
        """Convert a controlled operation into a tensor network in the pure
        state picture, but with only one open edge for each controlling qubit.

        A qubit in an operation can be regarded as a controlling qubit if its
        value in the computational basis is never changed by the operation. As
        such, its input wire and output wire can be represented by the same edge
        in a tensor network, thus simplifying the tensor network. In other
        words, the :attr:`tensor_pure` for a controlled operation will be a
        block diagonal matrix, and its :attr:`tensor_control` will be a more
        compact representation of the same matrix.

        As a special case, if the operation is a diagonal gate, then every qubit
        can be regarded as a controlling qubit. See :class:`Diagonal`.
        """
        from .converter import Converter
        return Converter.convert_control(self)

    def adjoint(self):  # pragma: no cover
        """Return the adjoint of a quantum operation. ``~op`` is an alias of
        ``op.adjoint()``.
        """
        raise NotImplementedError()

    def __invert__(self):
        return self.adjoint()


class ImmutableOperation(Operation):
    """Class for quantum operations with explicit tensor representations. The
    operation is not supposed to be modified.

    :param data: A tensor representation of the operation. For this base class,
        it will be the tensor returned by :attr:`~Operation.tensor_density`.
        This means that for an operation with :math:`n` input qubits and
        :math:`m` output qubits, the tensor should be of rank :math:`2(n+m)`.
        A derived class of this may use different representation more suitable
        for a specific class of operations.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar ~.shape: A list of strings describing the input and output qubits of
        the quantum operation, with each string describing a qubit involved. For
        this base class, each string should be one of "i", "o" or "io",
        indicating whether each qubit is an input qubit, an output qubit, or
        both.
    :vartype ~.shape: List[str]

    :ivar name: Name of the quantum operation. Defaults to "ImOp".
    :vartype name: str, optional
    """

    def __init__(self,
                 data: np.ndarray,
                 shape: List[str],
                 name="ImOp") -> None:
        Operation.__init__(self, name)
        self.shape = shape
        self.process(data)

    def __str__(self) -> str:  # pragma: no cover
        return str(vars(self))

    def process(self, data):
        """Convert the input data into an appropriately shaped tensor, and
        initialize the operation with this tensor.

        A derived class that uses a different tensor representation of the
        operation should override this function in order to do shape checking
        and initialization correctly.

        :param data: The ``data`` parameter passed to
            :meth:`ImmutableOperation.__init__`.
        :type data: acqdp.tensor_network.tensor_valued.TensorValued or
            np.ndarray
        """
        _, lin = self._indices_with_property("^i")
        _, lout = self._indices_with_property(".*o$")
        if isinstance(data, TensorValued):
            if data.shape != tuple([2] * (2 * (lin + lout))):
                raise ValueError("Invalid operation: Input dimensions does not match the claimed shape")
            self._tensor_density = data
        elif type(data) != np.ndarray or np.size(data) != np.prod(np.shape(data)):
            raise TypeError("Invalid operation: Operation should be an np.ndarray.")
        else:
            try:
                self._tensor_density = Tensor(np.reshape(data, tuple([2] * (2 * (lin + lout)))))
            except ValueError:
                raise ValueError("Invalid operation: Operation dimension does not match qubits.")

    @classmethod
    def operation_from_kraus(cls, shape, kraus, name="ImOp_Kraus"):
        """Construct a quantum operation from its Kraus operator representation.

        :param shape: The shape of the operation, in the same format as the
            :attr:`shape` attribute of :class:`ImmutableOperation`.
        :type shape: List[str]

        :param kraus: The list of Kraus operators.
        :type kraus: List[acqdp.tensor_network.tensor_valued.TensorValued or
            np.ndarray]

        :param name: Name of the resulting quantum operation. Defaults to
            "ImOp_Kraus".
        :type name: str, optional

        :returns: A quantum operation constructed from the given Kraus operator
            representation.
        :rtype: ImmutableOperation

        """
        lin = len([i for i in shape if i[0] == 'i'])
        lout = len([i for i in shape if i[-1] == 'o'])
        shape_tensor = tuple([2] * (lin + lout))
        tensor_density = np.zeros(tuple([2] * (2 * (lin + lout))))
        for operator in kraus:
            if not isinstance(operator, np.ndarray):
                raise TypeError("Invalid operation: Kraus Operator should be an np.ndarray.")
            if (operator.shape != shape_tensor) and (operator.shape != (2 ** lout, 2 ** lin)):
                raise ValueError("Invalid operation: Dimensions do not match")
            i = np.reshape(operator, shape_tensor)
            tensor_density += np.multiply.outer(i, np.conj(i))
        return cls(tensor_density, shape, name)

    @property
    def is_pure(self):
        """Return True if the operation is a pure operation.

        Pure operations include pure states, isometries, projections and their
        combinations.

        Note that currently this function determines whether an operation is
        pure solely based on the class of the operation, without inspecting the
        actual data. For example, if a quantum state ``s`` is initialized with
        ``s = State(num_qubits, data)``, then ``s.is_pure`` will always be False
        and ``s.tensor_pure`` will always raise an error, even if ``data`` is
        actually the density matrix of a pure state.
        """
        return isinstance(self, PureOperation)


class State(ImmutableOperation):
    """Class for simple quantum states.

    Quantum states are regarded as a special case of quantum operations where
    there is no input qubit, and each qubit is an output qubit.

    :param num_qubits: Number of qubits in the state.
    :type num_qubits: int

    :param data: The density matrix representation of the state.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum state. Defaults to "State".
    :vartype name: str, optional

    """

    def __init__(self,
                 num_qubits,
                 data: np.ndarray,
                 name="State") -> None:
        ImmutableOperation.__init__(self, data, tuple(['o'] * num_qubits), name)


class Measurement(ImmutableOperation):
    """Class for simple (destructive) measurements.

    Destructive measurements are regarded as a special case of quantum
    operations where each qubit is an input qubit, and there is no output
    qubit. Such a measurement maps an arbitrary quantum state to a number.

    :param num_qubits: Number of qubits measured.
    :type num_qubits: int

    :param data: The POVM (positive operator-valued measure) representation of
        the measurement.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum state. Defaults to "Meas".
    :vartype name: str, optional
    """

    def __init__(self,
                 num_qubits,
                 data: np.ndarray,
                 name="Meas") -> None:
        ImmutableOperation.__init__(self, data, tuple(['i'] * num_qubits), name)


class Channel(ImmutableOperation):
    """Class for simple quantum channels.

    This class is used for the common case where the input and output Hilbert
    spaces are the same, i.e., each qubit is both an input qubit and an output
    qubit. For channels that do not satisfy this constraint, please use
    :class:`ImmutableOperation` directly.

    :param num_qubits: Number of qubits the channel operates on.
    :type num_qubits: int

    :param data: The tensor representation of the channel in the density matrix
        picture.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum operation. Defaults to "Channel".
    :vartype name: str, optional
    """

    def __init__(self,
                 num_qubits,
                 data: np.ndarray,
                 name='Channel') -> None:
        ImmutableOperation.__init__(self, data, tuple(['io'] * num_qubits), name)


class PureOperation(ImmutableOperation):
    """Class for simple pure quantum operations.

    :param data: The tensor representation of the operation in the pure state
        picture. It will be the tensor returned by
        :attr:`~Operation.tensor_pure`. This means that for an operation with
        :math:`n` input qubits and :math:`m` output qubits, the tensor should be
        of rank :math:`n+m`, i.e., half the rank of the tensor in the density
        matrix picture.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar ~.shape: The shape of the operation, in the same format as the
        :attr:`shape` attribute of :class:`ImmutableOperation`.
    :vartype ~.shape: List[str]

    :ivar name: Name of the quantum operation. Defaults to "PureOp".
    :vartype name: str, optional

    :ivar self_adjoint: Whether the operation is self-adjoint. Defaults to
        False.
    :vartype self_adjoint: bool, optional
    """

    def __init__(self,
                 data: np.ndarray,
                 shape: List[str],
                 name="PureOp",
                 self_adjoint=False) -> None:
        ImmutableOperation.__init__(self, data, shape, name)
        self.self_adjoint = self_adjoint

    def set_adjoint_op(self, adjoint_op):
        """Set the known adjoint of the operation.

        Usually, the adjoint of a pure operation is constructed on demand by
        calculating the :attr:`~Operation.tensor_pure` of the operation, then
        constructing a new :class:`PureOperation` object, which can be an
        inefficient process. By setting the adjoint of a operation to a known
        value, one can bypass this procedure. Note that, for efficiency, there
        is no check that ``adjoint_op`` is actually the adjoint of ``self``.

        The adjoint of ``adjoint_op`` is also set to ``self`` so there is no
        need to use this function twice.

        :param adjoint_op: The known adjoint of ``self``.
        :type adjoint_op: PureOperation
        """
        self.adjoint_op = adjoint_op
        adjoint_op.adjoint_op = self

    def process(self, data):
        _, lin = self._indices_with_property("^i")
        _, lout = self._indices_with_property(".*o$")
        if isinstance(data, TensorValued):
            if data.shape != tuple([2] * (lin + lout)):
                raise ValueError("Invalid operation: Input dimensions does not match the claimed shape")
            self._tensor_pure = data
        elif type(data) != np.ndarray or np.size(data) != np.prod(np.shape(data)):
            raise TypeError("Invalid operation: Operation should be an np.ndarray.")
        else:
            try:
                self._tensor_pure = Tensor(np.reshape(data, tuple([2] * (lin + lout))))
            except ValueError:
                raise ValueError("Invalid operation: Operation dimension does not match qubits.")

    def adjoint(self):
        if self.self_adjoint:
            return self
        elif hasattr(self, 'adjoint_op'):
            return self.adjoint_op
        else:
            transition_dict = {'i': 'o', 'o': 'i'}
            shape = ["".join([transition_dict[i] for i in s[::-1]]) for s in self.shape]
            _, lin = self._indices_with_property("^i")
            _, lout = self._indices_with_property(".*o$")
            tensor_pure = (~self._tensor_pure) % tuple(list(range(lout, lout + lin)) + list(range(lout)))
            return PureOperation(tensor_pure, shape, "~" + self.name)


class PureState(PureOperation, State):
    """Class for simple pure quantum states.

    :param num_qubits: Number of qubits in the state.
    :type num_qubits: int

    :param data: The state vector representation of the state.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum state. Defaults to "PureState".
    :vartype name: str, optional
    """

    def __init__(self,
                 num_qubits,
                 data: np.ndarray,
                 name='PureState') -> None:
        PureOperation.__init__(self, data, tuple(['o'] * num_qubits), name, False)


class PureMeas(PureOperation, Measurement):
    """Class for simple projective measurements.

    :param num_qubits: Number of qubits measured.
    :type num_qubits: int

    :param data: The vector representation of the measurement. Note that it
        should be the *complex conjugation* of the state vector of the state
        projected onto.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum state. Defaults to "PureMeas".
    :vartype name: str, optional
    """

    def __init__(self,
                 num_qubits,
                 data: np.ndarray,
                 name='PureMeas') -> None:
        PureOperation.__init__(self, data, tuple(['i'] * num_qubits), name, False)


class Unitary(PureOperation, Channel):
    """Class for simple unitary gates.

    :param num_qubits: Number of qubits the unitary operates on.
    :type num_qubits: int

    :param data: The matrix representation of the unitary.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum state. Defaults to "Unitary".
    :vartype name: str, optional

    :ivar self_adjoint: Whether the unitary is self-adjoint. Defaults to False.
    :vartype self_adjoint: bool, optional
    """

    def __init__(self,
                 num_qubits,
                 data: np.ndarray,
                 name='Unitary',
                 self_adjoint=False) -> None:
        PureOperation.__init__(self, data, tuple(['io'] * num_qubits), name, self_adjoint)


class ControlledOperation(PureOperation):
    """Class for simple controlled operations.

    :param data: The tensor representation of the controlled operation. It will
        be the tensor returned by :attr:`~Operation.tensor_control`. This means
        that for an operation with :math:`k` controlling qubits, :math:`n`
        non-control input qubits and :math:`m` non-control output qubits, the
        tensor should be of rank :math:`k+n+m`.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar ~.shape: The shape of the operation, in the same format as the
        :attr:`shape` attribute of :class:`ImmutableOperation`, but in addition
        to "i", "o", and "io", the string "c" is also allowed, which indicates a
        controlling qubit.
    :vartype ~.shape: List[str]

    :ivar name: Name of the quantum operation. Defaults to "C-Op".
    :vartype name: str, optional

    :ivar self_adjoint: Whether the operation is self-adjoint. Defaults to
        False.
    :vartype self_adjoint: bool, optional
    """
    def __init__(self,
                 data: np.ndarray,
                 shape: List[str],
                 name="C-Op",
                 self_adjoint=False) -> None:
        PureOperation.__init__(self, data, shape, name, self_adjoint)

    def process(self, data):
        _, lctrl = self._indices_with_property("^c$")
        _, lin = self._indices_with_property("^i")
        _, lout = self._indices_with_property(".*o$")
        if isinstance(data, TensorValued):
            if data.shape != tuple([2] * (lctrl + lin + lout)):
                raise ValueError("Invalid operation: Input dimensions does not match the claimed shape")
            self._tensor_control = data
        elif type(data) != np.ndarray or np.size(data) != np.prod(np.shape(data)):
            raise TypeError("Invalid operation: Operation should be an np.ndarray.")
        else:
            try:
                self._tensor_control = Tensor(np.reshape(data, tuple([2] * (lctrl + lin + lout))))
            except ValueError:
                raise ValueError("Invalid operation: Operation dimension does not match qubits.")

    def adjoint(self):
        if self.self_adjoint:
            return self
        else:
            transition_dict = {'i': 'o', 'o': 'i', 'c': 'c'}
            shape = ["".join([transition_dict[i] for i in s[::-1]]) for s in self.shape]
            _, lctrl = self._indices_with_property("^c$")
            _, lin = self._indices_with_property("^i")
            _, lout = self._indices_with_property(".*o$")
            tensor_control = ~(self._tensor_control)\
                % tuple(list(range(lctrl))
                        + list(range(lctrl + lout, lctrl + lout + lin))
                        + list(range(lctrl, lctrl + lout)))
            return ControlledOperation(tensor_control, shape, "~" + self.name)


class Diagonal(ControlledOperation):
    """Class for simple diagonal gates.

    A diagonal gate can be regarded as a controlled phase shift, where every
    qubit is a controlling qubit.

    :param num_qubits: Number of qubits the diagonal gate operates on.
    :type num_qubits: int

    :param data: The diagonal elements of the matrix representation of the gate.
    :type data: acqdp.tensor_network.tensor_valued.TensorValued or np.ndarray

    :ivar name: Name of the quantum state. Defaults to "Diag".
    :vartype name: str, optional

    :ivar self_adjoint: Whether the gate is self-adjoint. Defaults to False.
    :vartype self_adjoint: bool, optional
    """
    def __init__(self,
                 num_cbits,
                 data: np.ndarray,
                 name='Diag',
                 self_adjoint=False) -> None:
        ControlledOperation.__init__(self, data, tuple(['c'] * num_cbits), name, self_adjoint)


XGate = Unitary(1, np.array([[0, 1], [1, 0]]), "X", True)
"""Single-qubit Pauli X gate."""

YGate = Unitary(1, np.array([[0, 1j], [-1j, 0]]), "Y", True)
"""Single-qubit Pauli Y gate."""

ZGate = Diagonal(1, np.array([1, -1]), "Z", True)
"""Single-qubit Pauli Z gate."""

HGate = Unitary(1, np.sqrt(0.5) * np.array([[1, 1], [1, -1]]), "H", True)
"""Single-qubit Hadamard gate."""

TGate = Diagonal(1, np.array([1, (1 + 1j) * np.sqrt(0.5)]), "T")
"""Single-qubit T gate, i.e., a :math:`\\pi/4` rotation around the Z axis on the Bloch sphere."""

SGate = Diagonal(1, np.array([1, 1j]), "S")
"""Single-qubit S gate, i.e., a :math:`\\pi/2` rotation around the Z axis on the Bloch sphere."""

Trace = Measurement(1, TensorNetwork(open_edges=[0, 0], bond_dim=2), "Tr")
"""The partial trace operation, defined as an measurement that maps every
normalized single-qubit state to 1.

In a circuit, a partial trace operation can simulate discarding a qubit. Note
that this operation is inherently not pure since discarding one part of an
entangled pure state will result in a mixed state.
"""


IGate = Unitary(1,
                TensorNetwork(open_edges=[0, 0], bond_dim=2),
                "I", True)
"""Single-qubit Identity gate."""


SWAPGate = Unitary(2,
                   TensorNetwork(open_edges=[0, 1, 1, 0], bond_dim=2),
                   "SWAP", True)
"""Two-qubit SWAP gate."""

CZGate = Diagonal(2, np.array([[1, 1], [1, -1]]), name='CZ', self_adjoint=True)
"""Two-qubit CZ gate."""

XHalfGate = Unitary(1,
                    np.array([[1 + 1j, 1 - 1j], [1 - 1j, 1 + 1j]]) * 0.5,
                    "X/2")
"""Single-qubit :math:`\\sqrt{X}` gate."""

YHalfGate = Unitary(1,
                    np.array([[1 + 1j, -1 - 1j], [1 + 1j, 1 + 1j]]) * 0.5,
                    "Y/2")
"""Single-qubit :math:`\\sqrt{Y}` gate."""

CNOTGate = ControlledOperation(np.array([[[1, 0], [0, 1]], [[0, 1], [1, 0]]]),
                             ['c', 'io'],
                             "CNOT", True)
"""Two-qubit CNOT gate."""

ZeroState = PureState(1, np.array([1, 0]), '|0>')
"""Single-qubit state :math:`|0\\rangle`."""

OneState = PureState(1, np.array([0, 1]), '|1>')
"""Single-qubit state :math:`|1\\rangle`."""

CompState = [ZeroState, OneState]
"""List of the computational basis states, :math:`|0\\rangle` and :math:`|1\\rangle`."""

PlusState = PureState(1, np.array([1/np.sqrt(2), 1/np.sqrt(2)]), '|+>')
"""Single-qubit state :math:`|+\\rangle`."""

MinusState = PureState(1, np.array([1/np.sqrt(2), -1/np.sqrt(2)]), '|->')
"""Single-qubit state :math:`|-\\rangle`."""

FourierState = [PlusState, MinusState]
"""List of the Hadamard basis states, :math:`|+\\rangle` and :math:`|-\\rangle`."""

ZeroMeas = PureMeas(1, np.array([1, 0]), '<0|')
"""Single-qubit measurement :math:`\\langle 0|`, the adjoint of :math:`|0\\rangle`."""

OneMeas = PureMeas(1, np.array([0, 1]), '<1|')
"""Single-qubit measurement :math:`\\langle 1|`, the adjoint of :math:`|1\\rangle`."""

CompMeas = [ZeroMeas, OneMeas]
"""List of the computational basis measurements, :math:`\\langle 0|` and :math:`\\langle 1|`."""

PlusMeas = PureMeas(1, np.array([1 / np.sqrt(2), 1 / np.sqrt(2)]), '<+|')
"""Single-qubit measurement :math:`\\langle +|`, the adjoint of :math:`|+\\rangle`."""

MinusMeas = PureMeas(1, np.array([1 / np.sqrt(2), -1 / np.sqrt(2)]), '<-|')
"""Single-qubit measurement :math:`\\langle -|`, the adjoint of :math:`|-\\rangle`."""

FourierMeas = [PlusMeas, MinusMeas]
"""List of the computational basis measurements, :math:`\\langle +|` and :math:`\\langle -|`."""

NDCompMeas = Channel(1, TensorNetwork([0,0,0,0], bond_dim=2), 'ND')
"""A non-destructive computational basis measurement.

This is equivalent to :class:`acqdp.circuit.noise.Dephasing()`, but with a
slightly more efficient tensor network representation.
"""

ZeroState.set_adjoint_op(ZeroMeas)
OneState.set_adjoint_op(OneMeas)
PlusState.set_adjoint_op(PlusMeas)
MinusState.set_adjoint_op(MinusMeas)


class XRotation(Unitary):
    """Class for X Rotation gates in the form of :math:`e^{i\\theta X}`.

    Note that, since the eigenvalues of the Pauli operator :math:`X` are 1 and
    -1, this actually corresponds to a rotation by :math:`2\\theta` on the Bloch
    sphere. For example, ``XRotation(np.pi)`` is a global phase operation with
    no physical effect, equivalent to a rotation by :math:`2\\pi` on the Bloch
    sphere. On the other hand, ``XRotation(np.pi / 2)`` is equivalent to the
    Pauli X operator up to a global phase, i.e., a rotation by :math:`\\pi`
    around the X axis on the Bloch sphere.

    :ivar angle: The rotation angle :math:`\\theta`.
    :vartype angle: float
    """
    def __init__(self, angle):
        self.angle = angle
        Unitary.__init__(self, 1,
                         np.array([[np.cos(angle), 1j * np.sin(angle)],
                                  [1j * np.sin(angle), np.cos(angle)]]),
                         "R_X")

    def adjoint(self):
        return XRotation(-self.angle)


class ZRotation(Diagonal):
    """Class for Z Rotation gates in the form of :math:`e^{i\\theta Z}`.

    The same note about :class:`XRotation` applies here. Furthermore, note that
    the rotation direction is the reverse of the "natural" direction given by
    :math:`\\mathop{\\mathrm{diag}}(1, e^{i\\theta})`, since the actual rotation
    is :math:`\\mathop{\\mathrm{diag}}(e^{i\\theta}, e^{-i\\theta})`. For
    example, ``ZRotation(-np.pi / 8)`` and ``ZRotation(-np.pi / 4)`` will give
    the T gate and the S gate respectively, up to global phases.

    :ivar angle: The rotation angle :math:`\\theta`.
    :vartype angle: float
    """
    def __init__(self, angle):
        self.angle = angle
        Diagonal.__init__(self, 1,
                          np.array([np.exp(1j * angle), np.exp(-1j * angle)]),
                          "R_Z({})".format(angle))

    def adjoint(self):
        return ZRotation(-self.angle)


class Circuit(Operation):
    """Class for quantum circuits that can be manipulated by adding and removing
    operations.

    Each operation in a circuit has a name that is unique within the circuit,
    not necessarily related to its own :attr:`name` attribute. Each operation
    also has a time step which determines the order of the operations. For
    applications where those do not matter, they can be kept at their default
    values; see :meth:`append` for those default values.

    A circuit is always empty when initialized.

    :ivar name: Name of the quantum circuit. Defaults to "Circuit".
    :vartype name: str, optional

    :ivar operations_by_name: A dict mapping from operation names to dicts
        describing operations in the circuit. Each of the latter dicts ``d`` has
        three entries:

        * ``d["operation"]`` is the actual :class:`Operation`.
        * ``d["time_step"]`` gives the time step at which the opration happens.
        * ``d["qubits"]`` is a list of qubits indicating which qubits in the
          circuit the operation is applied to.
    :vartype operations_by_name: Dict[hashable, dict]
    """
    def __init__(self, name='Circuit') -> None:
        Operation.__init__(self, name)
        self.operations_by_name = {}

    @property
    def max_time(self) -> float:
        """Return the maximum time step among all operations in the circuit."""
        if(len(self.operations_by_name) == 0):
            return -1
        return max(map(lambda x: x["time_step"], self.operations_by_name.values()))

    def append(self,
               operation: Operation,
               qubits: List[int],
               time_step: Optional[float] = None,
               name=None) -> None:
        """Add an operation to the circuit.

        If a time step is given, the operation is inserted into the circuit at
        the given time. Otherwise, the operation is appended to the end of the
        circuit.

        :param operation: The operation to add to the circuit.
        :type operation: Operation

        :param qubits: The qubits the operation is applied to. Qubit names can
            be anything as long as they are hashable and comparable with each
            other. Any qubits not already in the circuit are added to the
            circuit.
        :type qubits: list

        :param time_step: The time step at which the operation happens. Defaults
            to :attr:`max_time` plus one, which ensures that the operation is
            appended to the end of the cirucit.
        :type time_step: int or float, optional

        :param name: The name of the operation in the circuit. By default, the
            name is a tuple generated from ``time_step`` and ``qubits``,
            ensuring its uniqueness within the circuit.
        :type name: hashable, optional
        """
        if time_step is None:
            time_step = 1 + self.max_time
        if name is None:
            name = (time_step,
                    tuple([q for q in qubits]))
        if (len(qubits) != len(set(qubits))
           or len(qubits) != len(operation.shape)):
            raise ValueError("Invalid qubits: qubits not match operation.")
        self.operations_by_name[name] = {"operation": operation,
                                         "time_step": time_step,
                                         "qubits": qubits}
        return self

    @property
    def operations_by_time(self):
        """Return an :class:`OrderedDict` of dicts, which is an reorganization
        of :attr:`operations_by_name` such that ``c.operations_by_time[t][name]
        == c.operations_by_name[name]``, where ``t`` is the time step of the
        operation in question.

        The returned ``OrderedDict`` in ordered by increasing time.
        """
        time_steps = set(i["time_step"] for i in self.operations_by_name.values())
        d = {t: {} for t in time_steps}
        for g in self.operations_by_name:
            d[self.operations_by_name[g]["time_step"]][g] = self.operations_by_name[g]
        return collections.OrderedDict(sorted(d.items()))

    def tree_string(self, indent=0) -> str:  # pragma: no cover
        indentation = indent * INDENT
        s = indentation + "Ops:\n"
        d = self.operations_by_time
        for t in d:
            for n in d[t]:
                s += indentation + INDENT + str(n) + "\n"
                s += d[t][n]['operation'].tree_string(indent + 1)
        s += indentation + "End of circuit\n"
        return s

    @property
    def all_qubits(self) -> List[int]:
        """Return a list of all qubits in the circuit."""
        s = []
        for g in self.operations_by_name:
            s += self.operations_by_name[g]["qubits"]
        try:
            return sorted(set(s))
        except TypeError:
            dic = {str(a): a for a in s}
            return [dic[t] for t in sorted(dic)]

    @property
    def shape(self) -> List[str]:
        """Return the shape of the circuit as an operation, in the same format
        as the :attr:`~ImmutableOperation.shape` attribute of
        :class:`ImmutableOperation`.

        Raises an :class:`ValueError` if the circuit is invalid. This can happen
        if a qubit is used multiple times at the same time step, or if the
        shapes of adjacent operations on the same qubit do not fit together,
        such as a quantum state followed by another quantum state.
        """
        shape = {x: "" for x in self.all_qubits}
        trans_dict = {'b': 'i', 'd': 'o', 'c': 'io', 'i': 'i', 'o': 'o'}
        last_time_used = {x: -math.inf for x in self.all_qubits}
        for time_step in self.operations_by_time:
            for gate_name in self.operations_by_time[time_step]:
                gate = self.operations_by_name[gate_name]
                for i, qubit in enumerate(gate['qubits']):
                    if last_time_used[qubit] >= time_step:
                        raise ValueError("Invalid connection: Qubit {} used multiple times at timestep {}".format(qubit, time_step))
                    last_time_used[qubit] = time_step
                    shape[qubit] += "".join([trans_dict[letter] for letter in gate['operation'].shape[i]])
        pattern = re.compile("^o?(io)*i?$")
        for qubit in shape:
            if pattern.match(shape[qubit]):
                shape[qubit] = shape[qubit][: 2 - len(shape[qubit]) % 2]
            else:
                raise ValueError("Invalid connection on qubit {}".format(qubit))
        return tuple(shape[qubit] for qubit in sorted(shape))

    @property
    def is_valid(self) -> bool:
        """
        Return True if the circuit is valid.

        See :attr:`shape` for ways a circuit can be invalid.
        """
        try:
            self.shape
            return True
        except ValueError:
            return False

    @property
    def is_pure(self) -> bool:
        """Return True if the operation is a pure operation.

        Pure operations include pure states, isometries, projections and their
        combinations.

        Similar to :attr:`ImmutableOperation.is_pure`, this does not analyze the
        actual action of the circuit, and only checks whether the component
        operations are all pure.
        """
        return self.is_valid and np.all([x["operation"].is_pure for x in self.operations_by_name.values()])

    def adjoint(self):
        if not self.is_pure:
            raise ValueError("ValueError: Noisy circuit does not have an adjoint")
        c = Circuit()
        for operation_name in self.operations_by_name:
            operation = self.operations_by_name[operation_name]['operation']
            qubits = self.operations_by_name[operation_name]['qubits']
            time_step = self.max_time - self.operations_by_name[operation_name]['time_step']
            c.append(~operation, qubits, time_step)
        return c

    def __copy__(self):
        c = Circuit()
        for operation_name in self.operations_by_name:
            operation = self.operations_by_name[operation_name]['operation']
            qubits = self.operations_by_name[operation_name]['qubits']
            time_step = self.operations_by_name[operation_name]['time_step']
            c.append(operation, qubits, time_step, operation_name)
        return c


class ControlledCircuit(ControlledOperation):
    """Class for quantum circuits controlled by a single qubit.

    :ivar ~_.circuit: The circuit being controlled.
    :vartype ~_.circuit: Circuit

    :ivar name: Name of the controlled circuit. Defaults to "C-<circuit>", where
        <circuit> is the name of the circuit being controlled.
    :vartype name: str, optional

    :ivar conditioned_on: Whether the circuit is applied when the controlling
        qubit is in the :math:`|1\\rangle` state or the :math:`|0\\rangle`
        state. Defaults to True, meaning that the circuit is applied when the
        controlling qubit is in the :math:`|1\\rangle` state.
    :vartype conditioned_on: bool, optional
    """
    def __init__(self,
                 circuit,
                 name=None,
                 conditioned_on=True):
        if name is None:
            name = "C-" + circuit.name
        self.shape = tuple(['c'] + list(circuit.shape))
        self.circuit = circuit
        self.conditioned_on = conditioned_on
        self.name = name

    def tree_string(self, indent=0):  # pragma: no cover
        indentation = indent * INDENT
        s = indentation + "Conditioned on:{}\n".format(self.conditioned_on)
        s += self.circuit.tree_string(indent + 1)
        s += indentation + "End of control\n"
        return s

    def adjoint(self):
        return ControlledCircuit(~(self.circuit), conditioned_on=self.conditioned_on)

    @property
    def is_valid(self):
        """
        Return True if the circuit is valid.

        See :attr:`Circuit.shape` for ways a circuit can be invalid.
        """
        return self.circuit.is_valid


def Controlled(operation, name=None, conditioned_on=True):
    """Return a controlled version of the given operation.

    When the given operation is not applied, the identity gate is applied, which
    means the operation must have the same input and output qubits. The
    operation also needs to be pure in order for quantum control to be
    well-defined, which means it must be a unitary. However, it can be either
    an :class:`ImmutableOperation` or a :class:`Circuit`.

    :param operation: The quantum operation being controlled.
    :type operation: Operation

    :param name: Name of the resulting controlled operation. Defaults to
        "C-<operation>", where <operation> is the operation being controlled.
    :type name: str, optional

    :param conditioned_on: Whether the operation is applied when the controlling
        qubit is in the :math:`|1\\rangle` state or the :math:`|0\\rangle`
        state. Defaults to True, meaning that the operation is applied when the
        controlling qubit is in the :math:`|1\\rangle` state.
    :type conditioned_on: bool, optional
    """
    if name is None:
        name = "C{}-".format("" if conditioned_on else 0) + str(operation.name)
    if not (operation.is_pure and all(x == 'io' or x == 'c' for x in operation.shape)):
        raise ValueError("Controlled unitary must be unitary")
    shape = tuple(['c'] + list(operation.shape))
    len_qubits = len(operation.shape)
    _, len_cbits = operation._indices_with_property("^c$")
    if isinstance(operation, Circuit):
        return ControlledCircuit(operation, name, conditioned_on)
    else:
        data = operation.tensor_control.contract()
        tn = TensorNetwork()
        for i in range(len_qubits):
            tn.open_edge(i, 2)
        for i in range(len_cbits, len_qubits):
            tn.open_edge(i)
        identity = tn.contract()
        if conditioned_on:
            return ControlledOperation(np.stack((identity, data)), shape, name)
        else:
            return ControlledOperation(np.stack((data, identity)), shape, name)


class SuperPosition(Operation):
    """Class for a superposition of multiple pure operations.

    :ivar operations: A list of operations in the superposition.
    :vartype operations: List[Operation]

    :ivar coefs: A list of coefficients, corresponding to each of the component
        operations. By default, all the coefficients are 1.
    :vartype coefs: List[float], optional

    :ivar name: Name of the superposition operation. Defaults to "SuperPos".
    :vartype name: str, optional

    """
    def __init__(self,
                 operations,
                 coefs=None,
                 name='SuperPos'):
        if coefs is None:
            coefs = np.ones(len(operations))
        if len(coefs) != len(operations):
            raise ValueError("Invalid operation: length of components does not match length of coefficients")
        Operation.__init__(self, name)
        input_indices = operations[0]._input_indices
        output_indices = operations[0]._output_indices
        shape = list(operations[0].shape)
        for op in operations:
            if (op._input_indices != input_indices)\
               or (op._output_indices != output_indices):
                raise ValueError("Invalid operation: components have incompatible shape")
            if not op.is_pure:
                raise ValueError("Invalid superposition: all components must be pure")
            for s in range(len(op.shape)):
                if shape[s] == 'c':
                    shape[s] = 'io'
        self.coefs = coefs
        self.operations = operations
        self.shape = tuple(shape)

    @property
    def is_pure(self):
        """Return True if the operation is a pure operation.

        A :class:`SuperPosition` object is always pure.
        """
        return True

    def adjoint(self):
        return SuperPosition([~op for op in self.operations],
                             np.conj(self.coefs),
                             "~" + str(self.name))


def XXRotation(angle):
    """Return a two-qubit XX rotation gate in the form of
    :math:`e^{i\\theta(X\\otimes X)}`, also known as an Ising gate.

    :param angle: The rotation angle :math:`\\theta`.
    :type angle: float
    """
    return SuperPosition([IGate * IGate, XGate * XGate],
                         [np.cos(angle), np.sin(angle) * 1j],
                         'R_XX({})'.format(angle))


def ZZRotation(angle):
    """Return a two-qubit ZZ rotation gate in the form of
    :math:`e^{i\\theta(Z\\otimes Z)}`.

    :param angle: The rotation angle :math:`\\theta`.
    :type angle: float
    """
    return Diagonal(2,
                    np.array([[np.exp(1j * angle), np.exp(-1j * angle)],
                              [np.exp(-1j * angle), np.exp(1j * angle)]]),
                    'R_ZZ({})'.format(angle))



import copy
import numpy
from .tensor_valued import TensorValued, DTYPE
class TensorView(TensorValued):
    """
    :class:`TensorView` is a subclass of :class:`TensorValued` representing unary operations over another :class:`TensorValued` object that preserves the shape of the tensor. Common examples include element-wise conjugation and normalization with respect to the frobenius norm.

    :ivar tn: the underlying `TensorValued` object where the unary operation is performed onto.
    :ivar func: the unary function to be applied.
    :ivar homomorphism: indicator whether the unary function is homomorphic to the addition and multiplication of tensors. If so, the unary function can be broadcast to lower-level tensors, enabling potential simplification of the tensor network structure.
    :ivar dtype: dtype for the tensor entries.
    """
    def __init__(self, tn, func=numpy.conj, homomorphism=False, dtype=DTYPE):
        """
        The constructor of a `TensorView` object.
        """
        super().__init__(dtype)
        self.ref = tn
        self.func = func
        self.homomorphism = (func == numpy.conj) | homomorphism

    def __str__(self) -> str:
        data_str = "Data: \n" + str(self.ref)
        func_str = "Func: " + str(self.func)
        return super().__str__() + "\n" + data_str + func_str

    @property
    def shape(self):
        """
        The common property of all :class:`TensorValued` classes, yielding the shape of the object.
        :class:`TensorValued` objects must have compatible shapes in order to be connected together in a :class:`TensorNetwork`,
        or summed over in a :class:`TensorSum`.

        For :class:`TensorView` objects, it refers to the shape of the underlying :class:`TensorValued` object where the unary operation is performed onto.
        """
        return self.ref.shape

    @property
    def is_ready(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the current :class:`TensorValued` object is ready for
        contraction, i.e. whether it semantically represents a tensor with a definite value.
        In the process of a program, not all :class:`TensorValued` objects need to be ready; however once the `data` property of a certain object is queried,
        such object must be ready in order to successfully yield an :class:`numpy.ndarray` object.

        For :class:`TensorView` objects, it is to indicate whether the underlying :class:`TensorValued` object where the unary operation is performed onto, is
        ready for contraction.
        """
        return self.ref.is_ready

    @property
    def is_valid(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the :class:`TensorValued` object is valid or not.
        In every step of a program, all existing :class:`TensorValued` object must be valid, otherwise an exception should be thrown out;
        this property is for double checking that the current :class:`TensorValued` object is indeed valid.

        For :class:`TensorView` objects, it is to indicate whether the underlying :class:`TensorValued` object where the unary operation is performed onto, is
        valid or not.
        """
        return self.ref.is_valid

    @property
    def raw_data(self):
        """
        The data of the underlying :class:`TensorValued` object where the unary operation is performed onto.
        """
        return self.ref.contract()

    def fix_index(self, index, fix_to=0):
        """
        Fix the given index to the given value. The object after the method would have the same type as the original one, with rank 1 smaller than
        the original.

        :param index: The index to fix.
        :type index: :class:`int`.
        :param fix_to: the value to assign to the given index.
        :type fix_to: :class:`int`.

        :returns:  :class:`TensorView` -- The :class:`TensorView` object after fixing the given index.
        """
        self.ref = self.ref.fix_index(index, fix_to)

    def expand(self, recursive=False):
        """
        Commute the unary operation with the underlying tensor network, when the unary operation is a homomorphism for tensor network contractions.
        """
        from acqdp.tensor_network import TensorNetwork
        if not self.homomorphism or not isinstance(self.ref, TensorNetwork):
            return self
        else:
            k = self.ref.copy()
            if recursive:
                k.expand(recursive=True)
        for node_name in k.nodes_by_name:
            k.update_node(node_name, TensorView(k.network.nodes[(0, node_name)]['tensor'], self.func, self.homomorphism))
        return k

    def cast(self, dtype):
        self.dtype = dtype
        self.ref = self.ref.cast(dtype)
        return self

    def contract(self, **kwargs):
        return self.func(self.ref.contract(**kwargs))

    def copy(self):
        return TensorView(self.ref.copy(), self.func)

    def __deepcopy__(self, memo):
        return TensorView(copy.deepcopy(self.ref), self.func)



import inspect
import copy
import numpy
import time
from multiprocessing import Pool
from functools import lru_cache
from acqdp import tensor_network

DEFAULT_ENGINE = 'numpy'


@lru_cache(maxsize=None)
def getDeployPlugin(pluginName):
    """
    plugin is just an installed python module that has "tensor_network" submodule
    """
    import importlib
    res = importlib.import_module(pluginName)
    return res


class Contractor:
    """
    Contractor class for tensor network contraction takes a :class:`ContractionTask` object and execute it sequentially.
    For :class:`NetworkContractionTask`, multi-processing is available for further accelarate the computation.

    :ivar backend: When set to `jax`, large tensor contractions will make use of the `jax` backend. `numpy.einsum` is used otherwise.
    :ivar exeEngine: Extension interface for other contraction backends. Set to `None` by default. When set to `parallel`, subtasks will be computed simultaneously.
    """

    def __init__(self, exeEngine=None, backend='default', dtype=complex, **kwargs):
        self.exeEngine = exeEngine
        self.backend = backend
        self.dtype = numpy.dtype(dtype)


    def execute(self, tasks, lst=None, **kwargs):
        """
        Execute a contraction task.

        :param tasks: The task to be executed.
        :type tasks: :class:`acqdp.ContractionScheme`
        :param lst: The list of subtasks to be executed. If set to `None`, all subtasks are executed and merged together.
        :type lst: :class:`List`

        :returns: :class:`numpy.ndarray` -- Final result expressed as a multi-dimensional array.
        """
        tasks._load_data()
        if lst is None:
            lst = range(tasks.length)
        engine = self.exeEngine
        if engine is None:
            engine = DEFAULT_ENGINE

        if engine == 'numpy':
            return tasks._merge(
                {i: self._execute(tasks[i], **kwargs) for i in lst})
        elif inspect.ismodule(engine):
            res = engine.tensor_networkService.contractorExecute(tasks)
            return res
        elif engine == "parallel":
            with Pool() as p:
                return tasks._merge(p.starmap(self._execute, [tasks[i] for i in lst]))
        elif engine.startswith("plugin:"):
            pluginName = engine[7:]
            plugin = getDeployPlugin(pluginName)
            res = plugin.tensor_networkService.contractorExecute(tasks)
            return res

    def _execute(self,
                      task,
                      track=False,
                      normalize=False,
                      cnt=None,
                      **kwargs):
        if cnt is None:
            commands = task.commands
        else:
            commands = task.commands[:cnt]
        output = task.output
        for command in commands:
            if track:
                print("Current Memory usage = {}".format(
                    self._track_memory(commands) + 4))
            operation = command[0]
            lhs = command[1]
            rhs = command[2]
            kwargs = command[3]
            try:
                if operation == 'f':
                    res = numpy.moveaxis(lhs[0][1], kwargs['fix_idx'],
                                         range(len(kwargs['fix_idx'])))[tuple(
                                             [a[0] for a in kwargs['fix_to']])]
                    rhs[0] = (lhs[0][0], res)
                else:
                    if operation == 'c':
                        init_norm = sum([l[0][0] for l in lhs])
                        if 'expr' in kwargs:
                            if self.backend == 'jax':
                                res = kwargs['expr'](*[l[0][1] for l in lhs], backend='jax')
                            else:
                                res = kwargs['expr'](*[l[0][1] for l in lhs])
                        else:
                            res = numpy.array(numpy.einsum(kwargs['subscripts'],
                                               *[l[0][1] for l in lhs]))
                    elif operation == 'n':
                        init_norm = 0
                        res = kwargs['func'](numpy.exp(lhs[0][0]) * lhs[0][1],
                                             **kwargs)
                    if normalize:
                        import numexpr as ne
                        norm = ne.evaluate(
                            'max(res.real ** 2 + res.imag ** 2)')**0.5
                        if norm == 0:
                            rhs[0] = (0, numpy.zeros(res.shape))
                        else:
                            res /= norm
                            rhs[0] = (numpy.log(norm) + init_norm, res)
                    else:
                        rhs[0] = (init_norm, res)
            except Exception as e:
                print(e)
                print(command)
                raise e
        if cnt is None:
            res = numpy.exp(output[0][0]) * output[0][1]
            return numpy.array(res)

    def _track_memory(self, commands):
        lst = []
        for command in commands:
            if command[0] == 'c':
                for k in command[1]:
                    for j in lst:
                        if k[0] is None or j is k[0][1]:
                            break
                    else:
                        lst.append(k[0][1])
            elif command[0] == 'f':
                k = command[1]
                for j in lst:
                    if k[0] is None or j is k[0][1]:
                        break
                else:
                    lst.append(k[0][1])
        return numpy.log2(max([a.size for a in lst]))


defaultContractor = Contractor()


_defaultContractor = None

def getDefault():
    global _defaultContractor
    if _defaultContractor is None:
        try:
            from acqdp.tensor_network import contractor
            _defaultContractor = contractor.defaultContractor
        except ImportError:
            _defaultContractor = Contractor()
    return _defaultContractor


def setDefault(aContractor: Contractor):
    global _defaultContractor
    _defaultContractor = aContractor


def contract(tn, **kwargs):
    return getDefault().contract(tn, **kwargs)



import copy
import numpy
from .tensor_valued import TensorValued, DTYPE


class Tensor(TensorValued):
    """

    A :class:`Tensor` is an array of numbers with multiple dimensions. The most basic examples of a :class:`Tensor` are a vector (1-dimensional arrays of numbers)
    and a matrix (2-dimensional arrays).

    In our implementation, :class:`Tensor` is a subclass of :class:`TensorValued`, where the value is stored in an `numpy.ndarray`.
    All other `TensorValued` represent operations over the :class:`Tensor` objects.

    :ivar _data: `numpy.ndarray` object representing the data corresponding to the tensor.

    """

    def __init__(self,
                 data: numpy.ndarray = None,
                 dtype: type = DTYPE) -> None:
        """
        Constructor of a :class:`Tensor` object.
        """
        if data is None:
            self._data = None
        elif isinstance(data, TensorValued):
            self._data = data.contract()
        elif isinstance(data, numpy.ndarray):
            self._data = data
            if dtype is None:
                dtype = self._data.dtype
        else:
            self._data = numpy.array(data)
            if dtype is None:
                dtype = self._data.dtype
        super().__init__(dtype)

    @property
    def shape(self):
        """
        The common property of all :class:`TensorValued` classes.
        The shape of a `TensorValued` object is the bond dimension for each of its indices.
        :class:`TensorValued` objects must have compatible shapes in order to be connected together in a :class:`TensorNetwork`,
        or summed over in a :class:`TensorSum`.

        For :class:`Tensor` objects, it refers to the shape of the underlying :class:`numpy.ndarray` object.
        """
        if self._data is None:
            return None
        elif isinstance(self._data, numpy.ndarray):
            return self._data.shape
        else:
            raise ValueError

    def __str__(self) -> str:
        s = None
        if isinstance(self._data, numpy.ndarray):
            s = numpy.around(self._data, decimals=3)
        data_str = "Data: \n" + str(s)
        return super().__str__() + "\n" + data_str

    def __repr__(self) -> str:
        return "Id: " + str(self.identifier) + self.__str__()

    def __iadd__(self, t):
        self._data += t.contract()
        return self

    @property
    def is_valid(self) -> bool:
        """
        For :class:`Tensor` objects, it is to indicate whether the underlying :class:`numpy.ndarray` object where the unary operation is
        performed onto, is valid or not.
        """
        return True

    @property
    def is_ready(self) -> bool:
        """
        The common property of all :class:`TensorValued` classes, indicating whether the current :class:`TensorValued` object is ready for
        contraction, i.e. whether it semantically represents a tensor with a definite value.
        In the process of a program, not all :class:`TensorValued` objects need to be ready; however once the `data` property of a certain
        object is queried,
        such object must be ready in order to successfully yield an :class:`numpy.ndarray` object.

        For :class:`Tensor` objects, it is to indicate whether the underlying :class:`numpy.ndarray` object where the unary operation is
        performed onto, is ready for contraction.
        """
        return self._data is not None

    @property
    def norm_squared(self):
        """
        Square of Frobenius norm of the underlying :class:`numpy.ndarray` object.
        """
        return numpy.linalg.norm(self._data.flatten()) ** 2

    def fix_index(self, index, fix_to=0):
        """
        Fix the given index to the given value. The object would have the same dtype as the original
        one, with rank 1 smaller than the original.

        :param index: The index to fix.
        :type index: :class:`int`.
        :param fix_to: the value to assign to the given index.
        :type fix_to: :class:`int`
        """
        if self._data is not None:
            self._data = numpy.moveaxis(self._data, index, 0)[fix_to]
        return self

    def contract(self, **kwargs):
        """
        :returns: :class:`numpy.ndarray` -- the value of the tensor whose value is stored in an :class:`numpy.ndarray` object.
        """
        return self._data

    def cast(self, dtype):
        """
        Return a copy of the `Tensor` object with updated underlying dtype.
        """
        return Tensor(numpy.array(self._data, dtype), dtype)

    def copy(self):
        """
        Return a copy of the `Tensor` object.
        """
        return Tensor(self._data, self.dtype)

    def __deepcopy__(self, memo):
        tn = Tensor(copy.deepcopy(self._data), dtype=self.dtype)
        return tn



from opt_einsum import helpers, paths
from collections import namedtuple
import sys
import os

try:
    from . import opt_einsum_paths, opt_einsum_helpers

    for name in opt_einsum_paths.__all__:
        setattr(paths, name, getattr(opt_einsum_paths, name))

    for name in opt_einsum_helpers.__all__:
        setattr(helpers, name, getattr(opt_einsum_helpers, name))
except ImportError:
    print('Cython modules for opt_einsum are not built. ACQDP will function normally, but contraction scheme finding may be slower. To build those modules, run:')
    print(f'    {os.path.basename(sys.executable)} -m pip install Cython')
    print(f'    {os.path.basename(sys.executable)} -m pip install --force-reinstall --no-deps acqdp')
    print(f'(or reinstall acqdp from whatever source you prefer)')
    print()



from acqdp.tensor_network.contraction import OrderCounter, ContractionCost
import opt_einsum
import copy
import numpy as np

ORDER_RESOLVER_KWARGS = {'optimal': 8, 'dp': 20, 'thres': 1e8}


class OrderResolver:
    """

    Interfaces for conveniently converting between different formats of contraction orders. Also serves as a converter from non-pairwise contraction orders to pairwise contraction orders.

    :ivar optimal: The maximum number of tensors in a single step of a non-pairwise order, for which brute-force search is used to reconstruct the pairwise order.
    :ivar dp: The maximum number of tensors in a single step of a non-pairwise order, for which dynamic programming is used to reconstruct the pairwise order.
    :ivar thres: The minimum cost of a non-pairwise contraction order via default path to invoke more advanced search for a pairwise order, such as the `dp` or `optimal` methods.

    """

    def __init__(self,
                 optimal=8,
                 dp=20,
                 thres=1e8,
                 **kwargs):
        self.optimal = optimal
        self.dp = dp
        self.thres = thres

    def order_to_contraction_scheme(self, tn, order):
        subscripts = self.order_to_subscripts(tn, order)
        cost = ContractionCost()
        counter = OrderCounter(order)
        new_order = []
        for o, i in zip(order, subscripts):
            try:
                res = self.subscript_to_path(*i)
                new_order += self.path_to_paired_order(o, res[0], counter)
                cost += res[1]
            except IndexError as e:
                print(o, i, res[0], res[1])
                raise e
        from acqdp.tensor_network import ContractionScheme
        return ContractionScheme(new_order, cost=cost)

    def order_to_path(self, tn, order):
        tn_copy = tn.copy()
        tn_copy.fix()
        edges_list = tn_copy.edges_by_name
        nodes_list = tn_copy.nodes_by_name
        lhs = [
            ''.join(
                opt_einsum.get_symbol(edges_list.index(e))
                for e in tn_copy.network.nodes[(0, n)]['edges'])
            for n in nodes_list
        ]
        rhs = ''.join(
            opt_einsum.get_symbol(edges_list.index(e))
            for e in tn_copy.open_edges)
        path = []
        for o in order:
            i, j = nodes_list.index(o[0][0]), nodes_list.index(o[0][1])
            path.append((i, j))
            nodes_list.pop(max(i, j))
            nodes_list.pop(min(i, j))
            nodes_list.append(o[1])
        return ','.join(lhs) + '->' + rhs, path, {
            opt_einsum.get_symbol(edges_list.index(e)): e for e in edges_list
        }

    def order_to_subscripts_tree(self, tn, order):
        from acqdp.tensor_network.contraction_tree import ContractionTree
        tree = ContractionTree.from_order(tn, order)
        return tree.full_subscripts, tree.full_order

    def order_to_subscripts(self, tn, order):
        tn_copy = tn.copy()
        subs = []
        for o in order:
            if o[1] == '#':
                break
            try:
                _, stn = tn_copy.encapsulate(o[0], stn_name=o[1])
                subs.append(stn.subscripts(o[0]))
            except Exception as e:
                print(e)
                raise e
        if len(order) > 0:
            subs.append(tn_copy.subscripts(order[-1][0]))
        return subs

    def print_order(self, tn, order):
        subs, order = self.order_to_subscripts_tree(tn, order)
        steps = []
        for i, sub in enumerate(subs):
            a = len(set(sub[0][0]).difference(sub[0][1]))
            b = len(set(sub[0][1]).difference(sub[0][0]))
            c = len(set(sub[0][1]).intersection(sub[0][0]).difference(sub[1]))
            d = len(set(sub[0][1]).intersection(sub[0][0]).intersection(sub[1]))
            steps.append(
                f'Step {i}: <{a+b+c+d}> {sub[0][0]}[{a} x {c} x {d}] * {sub[0][1]}[{c} x {b} x {d}] -> {sub[1]}[{a} x {b} x {d}] \n {order[i]}'
            )
        return '\n'.join(steps)

    def subscript_to_path(self, lhs, rhs, shapes):
        optimize = None
        try:
            path, path_info = opt_einsum.contract_path(','.join(lhs) + '->' + rhs,
                                                    *shapes,
                                                    shapes=True,
                                                    optimize='auto')
        except ValueError as e:
            print(','.join(lhs) + '->' + rhs, shapes)
            raise e
        if len(lhs) > 2 and path_info.opt_cost > self.thres:
            if len(lhs) < self.optimal:
                optimize = 'optimal'
            elif len(lhs) < self.dp:
                optimize = 'dp'
        if optimize is not None:
            path, path_info = opt_einsum.contract_path(','.join(lhs) + '->' + rhs,
                                                    *shapes,
                                                    shapes=True,
                                                    optimize=optimize)
        return path, ContractionCost(path_info.opt_cost,
                                     path_info.largest_intermediate)

    def path_to_paired_order(self, order, path, counter):
        lhs, rhs = copy.copy(order[0]), order[1]
        # print(lhs, rhs)
        new_order = []
        for i in path:
            try:
                if len(i) == 1:
                    return [[[lhs.pop(i[0])], rhs]]
                else:
                    new_order.append([[lhs[i[0]], lhs[i[1]]], counter.cnt])
                    lhs.pop(max(i[0], i[1]))
                    lhs.pop(min(i[0], i[1]))
                    lhs.append(new_order[-1][1])
            except IndexError as e:
                print(order[0], i)
                raise e
        new_order[-1][1] = rhs
        return new_order


defaultOrderResolver = OrderResolver()


class LocalOptimizer:
    """

    :class:`LocalOptimizer` takes a pairwise contraction tree and do local optimization on the tree. Note that a connected subgraph of the contraction tree represents a sequence of intermediate steps that take some (potentially intermediate) tensors as input, and outputs a later intermediate tensor. This sequence of steps can be optimized based on solely the shapes of the input and output tensors associated to the subgraph, without looking at the rest of the contraction tree. This allows local reorganization of the tree.

    Each iteration of local optimization consists of two steps: In the first step, the contraction tree is divided into non-overlapping subgraphs each up to a given size, with the internal connections in the subgraphs removed. The internal connections are then reconstructed in the second phase using accelerated optimum contraction tree finding approach. The iterations can be repeated by each time randomly selecting a different patch of subgraph division.

    :ivar size: The maximum number of nodes to be included in one subgraph.
    :ivar resolver: The resolver to help reconstruct the pairwise order from subgraph divisions.

    """

    def __init__(self,
                 size=13,
                 **kwargs):
        self.size = size
        self.resolver = OrderResolver(**kwargs.get('resolver_params', {}))

    def _flatten_order(self, tn, order, offset=0):
        tn_copy = tn.copy()
        tn_copy.fix()
        size_dic = {}
        cnt_offset = 0
        subs = self.resolver.order_to_subscripts(tn_copy, order)
        for i, o in enumerate(order):
            size_dic[o[1]] = len(subs[i][1])
            for j, k in enumerate(o[0]):
                size_dic[k] = len(subs[i][0][j])
        cnt = -1
        while -cnt < len(order):
            rhs_list = {}
            for i, o in enumerate(order[:cnt]):
                rhs_list[o[1]] = (i, o[0])
            o = order[cnt]
            lhs = o[0]
            if len(lhs) >= self.size:
                cnt -= 1
                continue
            # choose the largest tensor
            sizes = {k: size_dic[k] for k in o[0]}
            largest_list = sorted(sizes, key=lambda x: sizes[x])[::-1]
            for l in largest_list:
                if (l not in rhs_list) or (sizes[l] < 7) or (
                        len(rhs_list[l][1]) + len(lhs) > self.size):
                    continue
                if cnt_offset < offset:
                    cnt_offset += 1
                    continue
                else:
                    o[0].remove(l)
                    o[0] += rhs_list[l][1]
                    order.pop(rhs_list[l][0])
                    break
            else:
                cnt -= 1
        return order

    def optimize(self, tn, order, num_iter):
        for _ in range(num_iter):
            new_order = self._flatten_order(tn,
                                           order.order,
                                           offset=np.random.randint(
                                               self.size))
            order = self.resolver.order_to_contraction_scheme(tn, new_order)
        return order



from acqdp.tensor_network.local_optimizer import defaultOrderResolver, LocalOptimizer
from acqdp.tensor_network.contraction import ContractionCost, ContractionScheme
from multiprocessing import Pool
import copy
import numpy
import opt_einsum
import tqdm


class Slicer:
    """
    :class:`Slicer` finds slicing of an unsliced contraction scheme when called by the :class:`SlicedOrderFinder`.

    :ivar num_iter_before: Number of iterations of local optimization before slicing.
    :ivar num_iter_before: Number of iterations of local optimization in the middle of slicing.
    :ivar num_iter_before: Number of iterations of local optimization after slicing.
    :ivar max_num_slice: Maxmimum number of edges to be sliced. If set to -1, the constraint will be ignored.
    :ivar num_threads: Number of threads for multi-processing.
    :ivar slice_thres: Automatically slice edges that introduce an overhed below this threshold. Set to 0.02 by default.
    """
    def __init__(self,
                 num_iter_before=0,
                 num_iter_middle=20,
                 num_iter_after=100,
                 max_tw=29,
                 max_num_slice=-1,
                 num_threads=28,
                 slice_thres=.02,
                 **kwargs):
        self.num_iter_before = num_iter_before
        self.num_iter_middle = num_iter_middle
        self.num_iter_after = num_iter_after
        self.max_tw = max_tw
        self.max_num_slice = max_num_slice
        self.num_threads = num_threads
        self.slice_thres = slice_thres
        self.local_optimizer = LocalOptimizer(
            **kwargs.get('local_optimizer_params', {}))
        self.num_suc_candidates = kwargs.get('num_suc_candidates', 10)

    def _slice(self, tn, orders, num_process=0):
        tn = tn.copy()
        tnc = tn
        while True:
            try:
                tn = tnc.copy()
                y = min(orders, key=lambda a: (a.cost, orders.index(a)))
                slice_edges = []
                print(f'Process {num_process} initial cost: {y.cost}', flush=True)
                y = self.local_optimizer.optimize(
                    tn, y, self.num_iter_before)
                while y.cost.t > 2**self.max_tw:
                    y = self.local_optimizer.optimize(
                        tn, y, self.num_iter_middle)
                    k, order = self._biggest_weight_edge(tn, y.order)
                    slice_edges += k
                    for a in k:
                        tn.fix_edge(a)
                    tn.fix()
                    y = defaultOrderResolver.order_to_contraction_scheme(tn, order)
                    y.cost.k = len(slice_edges)
                    if len(slice_edges) >= self.max_num_slice:
                        break
                    if numpy.log2(float(y.cost.t)) - self.max_tw + len(
                            slice_edges) >= self.max_num_slice + 2: # early termination
                        break
                new_y = self.local_optimizer.optimize(
                    tn, y, self.num_iter_after)
                new_y.cost.k=len(slice_edges)
                if new_y.cost.t <= 2**self.max_tw:
                    y = new_y
                if y.cost.t <= 2**self.max_tw:
                    print(f'Process {num_process} succeeded with {y.cost}',
                          flush=True)
                    return ContractionScheme(y.order, slice_edges, cost=y.cost)
                else:
                    return None
            except KeyboardInterrupt:
                return None

    def _biggest_weight_edge(self, tn, order):
        """
        Find an edge or a list of edges, slicing of which introduces an overhead each that is below a threshold given by self.slice_thres, or a minimal overhead if self.slice_thres is unattainable.
        The method enumerates all edges that appear frequently on the stem of the contraction tree. It tries to introduce as minimal overhead as possible by flipping branches on the stem while trying to slice the edges.
        """
        tn_copy = tn.copy()
        tn_copy.fix()
        nodes_names = list(tn_copy.nodes_by_name)
        from acqdp.tensor_network.undirected_contraction_tree import UndirectedContractionTree
        eq, path, eedd = defaultOrderResolver.order_to_path(tn_copy, order)
        uct = UndirectedContractionTree(eq, path)
        se = set()
        edges_dic = {}
        ss = []
        for i in range(len(uct.stem) - 1):
            new_se = uct.open_subscripts_at_edge(
                uct.graph.nodes[uct.stem[i]]['parent'], uct.stem[i])
            for a in new_se.difference(se):
                edges_dic[a] = i
            for a in se.difference(new_se):
                ss.append((a, edges_dic[a], i))
            se = new_se
        ss = sorted(ss, key=lambda x: x[1] - x[2])[:10]
        ss_dic = {}
        c = uct.cost
        slice_edges = []
        for s in ss:
            uct_copy = copy.deepcopy(uct)
            for v in range(uct_copy.n):
                uct_copy.graph.nodes[v]['subscripts'] = uct_copy.graph.nodes[v][
                    'subscripts'].difference({s[0]})
            for u, v in uct_copy.graph.edges:
                uct_copy.graph[u][v].clear()
                uct_copy.graph[v][u].clear()
                uct_copy.preprocess_edge(u, v)
                uct_copy.preprocess_edge(v, u)
            uct_copy.compute_root_cost()
            for v in range(uct_copy.n, uct_copy.n * 2 - 2):
                uct_copy.compute_node_cost(v)
            res = (uct_copy.cost, s[1], s[2], uct_copy.get_path())
            i = s[1]

            curr_cost = res[0]
            while i > 5:
                ii = i
                while ii > 3:
                    for k in range(3, ii)[::-1]:
                        for l in range(k, ii):
                            uct_copy.switch_branches(l)
                        if uct_copy.cost <= curr_cost:
                            curr_cost = uct_copy.cost
                            ii -= 1
                            break
                        else:
                            for l in range(k, ii)[::-1]:
                                uct_copy.switch_branches(l)
                    else:
                        break

                k = ii
                sss = ii + 1
                while k > 5:
                    uct_copy.switch_branches(k)
                    if uct_copy.cost <= curr_cost:
                        curr_cost = uct_copy.cost
                        sss = k
                    k -= 1
                ii = sss
                for k in range(6, ii):
                    uct_copy.switch_branches(k)
                if ii >= i:
                    break
                i = ii
            res = (curr_cost, i, res[2])
            j = res[2]
            while j < len(uct_copy.stem) - 5:
                jj = j
                while jj < len(uct_copy.stem) - 3:
                    for k in range(jj + 1, len(uct_copy.stem) - 1):
                        for l in range(jj, k)[::-1]:
                            uct_copy.switch_branches(l)
                        ucost = uct_copy.cost
                        if ucost <= curr_cost:
                            curr_cost = ucost
                            jj += 1
                            break
                        else:
                            for l in range(jj, k):
                                uct_copy.switch_branches(l)
                    else:
                        break
                k = jj
                sss = jj - 1
                while k < len(uct_copy.stem) - 5:
                    uct_copy.switch_branches(k)
                    ucost = uct_copy.cost
                    if ucost <= curr_cost:
                        curr_cost = uct_copy.cost
                        sss = k
                    k += 1
                jj = sss
                for k in range(jj + 1, len(uct_copy.stem) - 5)[::-1]:
                    uct_copy.switch_branches(k)
                if jj <= j:
                    break
                j = jj
            if curr_cost < (1 + self.slice_thres) / 2 * c:
                slice_edges.append(eedd[s[0]])
                c = curr_cost
                uct = uct_copy
            else:
                ss_dic[s[0]] = (curr_cost, res[1], j, uct_copy.get_path())
        if len(slice_edges) > 0:
            pp = uct.get_path()
        else:
            kk = sorted(ss_dic, key=lambda x: ss_dic[x][0])[0]
            slice_edges = [eedd[kk]]
            pp = ss_dic[kk][-1]
        new_order = []
        for i, p in enumerate(pp):
            new_order.append([[nodes_names[p[0]], nodes_names[p[1]]], order[i][1]])
            nodes_names.pop(max(p[0], p[1]))
            nodes_names.pop(min(p[0], p[1]))
            nodes_names.append(order[i][1])
        return slice_edges, new_order

    def slice(self, tn, order_gen):
        orders = [next(order_gen) for _ in range(self.num_suc_candidates)]
        return self._slice(tn, orders)


def mpwrapper(slicer, tn, orders, num_process):
    return slicer._slice(tn, orders, num_process)


class MPSlicer(Slicer):
    """
    Multi-processing slicing, by concurrently trying different slicing routes.
    """
    def slice(self, tn, order_gen):
        candidates = []
        while len(candidates) <= self.num_suc_candidates:
            with Pool(self.num_threads) as p:
                lk = list((self, tn, [next(order_gen)], num_process) for num_process in range(self.num_threads))
                new_candidates = p.starmap(mpwrapper, lk)
            candidates += [i for i in new_candidates if i is not None]
            print("Num of candidates now: {}".format(len(candidates)))
        res = min(candidates, key=lambda x: (x.cost, candidates.index(x)))
        return res


def get_slicer(**kwargs):
    slicers = {'default': Slicer, 'mp': MPSlicer}
    slicer_name = kwargs.get('slicer_name', 'default')
    return (slicers[slicer_name])(**kwargs.get('slicer_params', {}))



import networkx as nx

class UndirectedContractionTree:
    """
    Contraction tree corresponding to a pairwise sequential contraction order, with interfaces for branch flipping and merging.
    TODO: Merge this class with :class:`ContractionTree` for a unified interface on contraction trees.
    """
    def __init__(self, eq, path):
        lhs, rhs = eq.replace(' ', '').split('->')
        lhs = lhs.split(',')
        self.n = len(lhs)
        self.open_subscripts = set(rhs)
        if self.n == 1:
            raise ValueError('Cannot construct undirected contraction tree with only one operand')
        self.graph = nx.Graph()
        operands = list(range(self.n))
        self.graph.add_nodes_from((i, {'subscripts': set(x)}) for i, x in enumerate(lhs))
        t = self.n
        for i, j in path[:-1]:
            if i > j: i, j = j, i
            v = operands.pop(j)
            u = operands.pop(i)
            self.graph.add_edge(u, t)
            self.graph.add_edge(v, t)
            self.graph.nodes[u]['parent'] = self.graph.nodes[v]['parent'] = t
            operands.append(t)
            t += 1
        u, v = self.root = tuple(operands)  # There should be two operands left
        self.graph.add_edge(u, v)
        self.graph.nodes[u]['parent'] = v
        self.graph.nodes[v]['parent'] = u
        for u, v in self.graph.edges:
            self.preprocess_edge(u, v)
            self.preprocess_edge(v, u)
        self.cost = self.root_cost = 0
        self.compute_root_cost()
        for v in range(self.n, self.n * 2 - 2):
            self.compute_node_cost(v)
        self.detect_stem()

    def is_leaf(self, v):
        return v < self.n

    def preprocess_edge(self, u, v):
        if v not in self.graph[u][v]:
            if self.is_leaf(v):
                self.graph[u][v][v] = {'subscripts': self.graph.nodes[v]['subscripts']}
            else:
                children = [c for c in self.graph[v] if c != u]
                assert len(children) == 2
                d0, d1 = (self.preprocess_edge(v, c) for c in children)
                self.graph[u][v][v] = {'subscripts': d0['subscripts'] | d1['subscripts']}
        return self.graph[u][v][v]

    def open_subscripts_at_edge(self, u, v):
        edge = self.graph[u][v]
        if self.is_leaf(v):
            return edge[v]['subscripts']
        return edge[v]['subscripts'] & (edge[u]['subscripts'] | self.open_subscripts)

    def compute_cost(self, s0, s1, s):
        res = 2 ** len(s0 | s1 | s)
        if (s0 | s1) - s:
            res *= 2
        return res

    def compute_root_cost(self):
        u, v = self.root
        s0 = self.open_subscripts_at_edge(u, v)
        s1 = self.open_subscripts_at_edge(v, u)
        self.cost -= self.root_cost
        self.root_cost = self.compute_cost(s0, s1, self.open_subscripts)
        self.cost += self.root_cost

    def compute_node_cost(self, v):
        parent = self.graph.nodes[v]['parent']
        children = [c for c in self.graph[v] if c != parent]
        assert len(children) == 2
        s0, s1 = (self.open_subscripts_at_edge(v, c) for c in children)
        if 'cost' in self.graph.nodes[v]: self.cost -= self.graph.nodes[v]['cost']
        self.graph.nodes[v]['cost'] = self.compute_cost(s0, s1, self.open_subscripts_at_edge(parent, v))
        self.cost += self.graph.nodes[v]['cost']

    def get_internal_path(self, v):
        if self.is_leaf(v):
            return []
        parent = self.graph.nodes[v]['parent']
        c0, c1 = [c for c in self.graph[v] if c != parent]
        return self.get_internal_path(c0) + self.get_internal_path(c1) + [(c0, c1, v)]

    def get_path(self):
        operands = list(range(self.n))
        res = []
        for r in self.root:  # Two halves of the contraction tree
            for c0, c1, v in self.get_internal_path(r):
                res.append((operands.index(c0), operands.index(c1)))
                operands.remove(c0)
                operands.remove(c1)
                operands.append(v)
        return res + [(0, 1)]

    def switch_edges(self, e0, e1):
        """
        Switch the edges e0 = (u, c0) and e1 = (v, c1) by connecting
        c0 to v and c1 to u instead. The path c0 -- u -- v -- c1 must
        be a path in the tree.
        This might break the stem, so if one or both of e0 and e1 may
        be on the stem, please use `switch_branches`, `merge_branches`,
        `unmerge_branches`, or fix the stem manually.
        """
        for (u, c0), (v, c1) in (e0, e1), (e1, e0):
            self.graph.add_edge(v, c0)
            self.graph[v][c0][c0] = self.graph[u][c0][c0]
            self.graph[v][c0][v] = self.graph[u][c0][u]
            self.graph.remove_edge(u, c0)
            if self.graph.nodes[c0]['parent'] == u:
                self.graph.nodes[c0]['parent'] = v
            if self.graph.nodes[u]['parent'] == c0:
                self.graph.nodes[v]['parent'] = c0
                self.graph.nodes[u]['parent'] = v
            if set(self.root) == {u, c0}: self.root = (v, c0)
        self.graph[u][v].clear()
        self.preprocess_edge(u, v)
        self.preprocess_edge(v, u)
        self.compute_node_cost(u)
        self.compute_node_cost(v)
        if set(self.root) == {u, v}: self.compute_root_cost()

    def step_root(self, u, v):
        if u not in self.root:
            raise ValueError('The new root must be adjacent to the old root at vertex u')
        self.root = u, v
        self.graph.nodes[u]['parent'] = v
        self.compute_node_cost(u)
        self.compute_root_cost()

    def detect_stem_in_subtree(self, u, v):
        """
        Detect the heaviest path in the subtree descending from the
        edge (u, v), *and* the heaviest such path that ends at v.
        Return format is (cost, endpoint, endpoint), (cost, endpoint).
        """
        if self.is_leaf(v):
            return (0, v, v), (0, v)
        c0, c1 = [c for c in self.graph[v] if c != u]
        cost_v = self.graph.nodes[v]['cost']
        res0, (cost0, end0) = self.detect_stem_in_subtree(v, c0)
        res1, (cost1, end1) = self.detect_stem_in_subtree(v, c1)
        cost, end = max((cost0, end0), (cost1, end1))
        res = max(res0, res1, (cost0 + cost1 + cost_v, end0, end1), key=lambda x: x[0])
        return res, (cost + cost_v, end)

    def detect_stem(self):
        # Just choose any leaf as the root to make things easier.
        (cost, end0, end1), _ = self.detect_stem_in_subtree(0, self.graph.nodes[0]['parent'])
        # Maybe not the most efficient, but convenient enough
        self.stem = nx.shortest_path(self.graph, end0, end1)
        return self.stem

    def switch_branches(self, i):
        """
        Switch the branches at nodes i and i+1 of the stem.
        """
        t0, u, v, t1 = self.stem[i - 1:i + 3]
        c0, = set(self.graph[u]) - {t0, v}
        c1, = set(self.graph[v]) - {u, t1}
        self.switch_edges((u, c0), (v, c1))

    def merge_branches(self, i):
        """
        Merge the branches at nodes i and i+1 of the stem.
        """
        t0, u, v, t1 = self.stem[i - 1:i + 3]
        c0, = set(self.graph[u]) - {t0, v}
        self.switch_edges((u, c0), (v, t1))
        self.stem.pop(i + 1)
        return c0

    def unmerge_branches(self, i, c0=None):
        """
        Split the branch at node i of the stem into two branches, with
        the sub-branch c0 now directly connecting to node i of the stem
        and the other sub-branch connecting to node i+1 of the stem. By
        default, use the sub-branch with smaller id as c0.
        """
        t0, u, t1 = self.stem[i - 1:i + 2]
        v, = set(self.graph[u]) - {t0, t1}
        if c0 is None:
            c0 = min(c for c in self.graph[v] if c != u)
        self.switch_edges((u, t1), (v, c0))
        self.stem.insert(i + 1, v)



import numpy
import copy
from collections import OrderedDict
from .tensor_valued import TensorValued, DTYPE


class TensorSum(TensorValued):
    """
    A :class:`TensorSum` object represents the summation of multiple tensors.

    :ivar terms_by_name: a dictionary with key-value pairs, where the key is the name of a summand and the value is the corresponding summand :class:`TensorValued` object.
    """
    def __init__(self, terms=None, dtype: type = DTYPE) -> None:
        """
        The constructor of a `TensorSum` object.
        """
        super().__init__(dtype)
        if terms is None:
            self.terms_by_name = OrderedDict()
        else:
            for term_name in self.terms_by_name:
                self.terms_by_name[term_name] = terms[term_name]

    def __str__(self):
        term_str = "\nTerms:"
        for term_name in self.terms_by_name:
            term_str += "\n" + str(term_name) + "\n" + str(self.terms_by_name[term_name])
        return super().__str__() + term_str

    def _update_shape(self, curr, tmp):
        if tmp is None:
            return curr
        if curr is None:
            return list(tmp)
        if len(curr) != len(tmp):
            raise ValueError('Component shapes do not match')
        for i in range(len(curr)):
            if curr[i] is None:
                curr[i] = tmp[i]
            elif (tmp[i] is not None) and (tmp[i] != curr[i]):
                raise ValueError('Component shapes do not match')
        return curr

    def _invalidate_shape_cache(self):
        if hasattr(self, '_cached_shape'):
            del self._cached_shape

    @property
    def shape(self):
        """
        The common property of all :class:`TensorValued` classes, yielding the shape of the object.
        :class:`TensorValued` objects must have compatible shapes in order to be connected together in a :class:`TensorNetwork`,
        or summed over in a :class:`TensorSum`.
        """
        if not hasattr(self, '_cached_shape'):
            curr = None
            for tsr_name in self.terms_by_name:
                tsr = self.terms_by_name[tsr_name]
                curr = self._update_shape(curr, tsr.shape)
            self._cached_shape = curr
        return tuple(self._cached_shape) if self._cached_shape is not None else None

    @property
    def is_valid(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the :class:`TensorValued` object is valid or not.
        In every step of a program, all existing :class:`TensorValued` object must be valid, otherwise an exception should be thrown out;
        this property is for double checking that the current :class:`TensorValued` object is indeed valid.
        """
        try:
            self.shape
        except ValueError:
            return False
        else:
            return True

    @property
    def is_ready(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the current :class:`TensorValued` object is ready for
        contraction, i.e. whether it semantically represents a tensor with a definite value.
        In the process of a program, not all :class:`TensorValued` objects need to be ready; however once the `data` property of a certain object is queried,
        such object must be ready in order to successfully yield an :class:`numpy.ndarray` object.
        """
        for t in self.terms_by_name.values():
            if not t.is_ready:
                return False
        return self.is_valid

    def add_term(self, term=None, tensor=None):
        """
        Add a term to the summation.

        :param term: Name of the term to be added. If not given, an auto-assigned one will be given as the output.
        :type term: hashable

        :param tensor: Value of the term to be added.
        :type tensor: :class:`TensorValued` or None

        :returns: The name of the newly added term.
        """
        from .tensor import Tensor
        if not isinstance(tensor, TensorValued):
            tensor = Tensor(tensor)
        if term is None:
            term = tensor.identifier
        if tensor.dtype == complex:
            self.dtype = complex
        if term in self.terms_by_name:
            raise KeyError("term {} to be added into the tensor network already in the tensor network!".format(term))
        if tensor.shape is not None:
            self.shape  # Make sure the shape cache is initialized
            self._cached_shape = self._update_shape(self._cached_shape, tensor.shape)
        self.terms_by_name[term] = tensor
        return term

    def __iadd__(self, t):
        self.add_term(tensor=t)
        return self

    def update_term(self, term, tensor=None):
        """
        Update the value of a term in the summation.

        :param term: Name of the term to be updated.
        :type term: hashable
        :param tensor: New value of the term
        :type tensor: :class:`TensorValued`

        :returns: Name of the term to be updated.
        """
        from .tensor import Tensor
        if (type(tensor) == numpy.ndarray) or (tensor is None):
            tensor = Tensor(tensor)
        if term not in self.terms_by_name:
            raise KeyError("term {} not in the TensorSum object".format(term))
        self.terms_by_name[term] = tensor
        self._invalidate_shape_cache()
        return term

    def remove_term(self, term):
        """
        Remove a term from the summation.

        :param term: Name of the term to be removed.
        :type term: hashable
        :returns: :class:`TensorValued` Value of the removed term
        """
        pop = self.terms_by_name.pop(term)
        self._invalidate_shape_cache()
        return pop

    def fix_index(self, index, fix_to=0):
        """
        Fix the given index to the given value. The result :class:`TensorValued` object would have the same type as the original one, with
        rank 1 smaller than the original.

        :param index: The index to fix.
        :type index: :class:`int`.
        :param fix_to: The value to assign to the given index.
        :type fix_to: :class:`int`.
        :returns:  :class:`TensorValued` -- The :class:`TensorValued` object after fixing the given index.
        :raises: NotImplementedError
        """
        ts = self.copy()
        for term in ts.terms_by_name:
            ts.terms_by_name[term] = ts.terms_by_name[term].fix_index(index, fix_to)
        ts._invalidate_shape_cache()
        return ts

    def cast(self, dtype):
        self.dtype = dtype
        for term in self.terms_by_name:
            self.update_term(term, self.terms_by_name[term].cast(dtype))
        return self

    def contract(self, **kwargs):
        """
        Evaluate the object by summing over all the terms.
        :returns: :class:`numpy.ndarray`
        """
        res = [
            self.terms_by_name[term].contract(**kwargs)
            for term in self.terms_by_name
        ]
        return sum(res)

    def copy(self):
        ts = TensorSum(dtype=self.dtype)
        for t in self.terms_by_name:
            ts.add_term(t, self.terms_by_name[t])
        return ts

    def __deepcopy__(self, memo):
        ts = TensorSum(dtype=self.dtype)
        for t in self.terms_by_name:
            ts.add_term(t, copy.deepcopy(self.terms_by_name[t]))
        return ts



import copy
import networkx
from collections import OrderedDict
from .tensor_valued import TensorValued, DTYPE
import numpy
import itertools
import opt_einsum
from scipy import sparse
from acqdp.utility import opt_einsum_ext
from functools import cmp_to_key
import json
import os

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class TensorNetwork(TensorValued):
    """
    A :class:`TensorNetwork` is a collection of :class:`Tensor`, with indices from different tensors identified and summed up to present a new tensor. An example of tensor network is two 2-tensors with a common edge representing the matrix multiplication. A :class:`TensorNetwork` is a graphical representation of an Einstein summation.

    :ivar network: attributed :class:`networkx.Graph` object, representing the attributed hypergraph of the tensor network as a tanner graph.
    :ivar dtype: dtype of the :class:`TensorValued` object.
    :ivar open_edges: a list of edge names, indicating the outgoing wires of the tensor network.
    """

    def __init__(self,
                 open_edges=None,
                 dtype: type = DTYPE,
                 bond_dim=None) -> None:
        """
        Constructor of :class:`TensorNetwork` class.
        """
        super().__init__(dtype)

        self.network = networkx.Graph()
        self.open_edges = []
        if open_edges is not None:
            for edge in open_edges:
                self.open_edge(edge, bond_dim)
        self.id_count = 0
        self.refresh_signature()

    def contract(self, preset=None, config=None, **kwargs):
        """
        Evaluate the :class:`TensorNetwork` object as an :class:`numpy.ndarray`.

        :param preset: Select a preset mode. If set to `default`, will contract the tensor network with default ordering and no post-processing. If set to `khp`, will invoke advanced KaHyPar-based contraction order finding routine. If set to `None`, will read config file or input keyword arguments for order finding specifications.
        :type preset: `"default"`, `"khp"` or `None`.
        :param config: File name for the config file. If preset option is not given and config file is given, the keyword arguments will be ignored and the specification will be read from the config file.
        :type config: `str`
        :param kwargs: If neither the preset option nor the config file name is given, the program will read the keyword arguments for order finding specifications. See :class:`OrderFinder`, :class:`Compiler`, :class:`Contractor` for individual specifications.
        :type kwargs: `dict`

        :returns: :class:`numpy.ndarray`.
        """

        if preset == 'default':
            config = os.path.join(THIS_DIR, 'default_params.json')
        elif preset == 'khp':
            config = os.path.join(THIS_DIR, 'khp_params.json')
        elif preset is not None:
            raise ValueError("Invalid preset options")

        if config is not None:
            with open(config, 'r') as f:
                kwargs = json.load(f)

        return self.compile(self.find_order(**kwargs), **kwargs).execute(**kwargs)


    def find_order(self, input_file=None, output_file=None, **kwargs):
        """
        Find a contraction scheme of the :class:`TensorNetwork` object. Equivalent to `next(get_order_finder(**kwargs).find_order(self))`. See :meth:`OrderFinder.find_order`.


        :param input_file: Input file name. When given, load the contraction scheme from the input file.
        :type input_file: str, optional
        :param output_file: Output file name. When given, the contraction scheme found will be dumped into the file.
        :type output_file: str, optional
        :returns: :class:`ContractionScheme`
        """
        from acqdp.tensor_network import ContractionScheme
        if input_file is not None:
            try:
                with open(input_file, 'r') as f:
                    return ContractionScheme.load(f)
            except FileNotFoundError:
                print('Order file not found. Proceed to order finding procedure')
        from acqdp.tensor_network.order_finder import get_order_finder
        res = next(get_order_finder(**kwargs).find_order(self))
        if output_file is not None:
            with open(output_file, 'w') as f:
                res.dump(f)
                print("Order result saved at "+ output_file)
        return res

    def compile(self, order, **kwargs):
        """
        Compile a :class:`ContractionScheme` corresponding to the :class:`TensorNetwork` object into a runtime executable contraction process as a :class:`ContractionTask`. Equivalent to `Compiler(**kwargs.get('compiler_params', {})).compile(self, order)` (See :meth:`Compiler.compile`).

        :returns: :class:`ContractionTask`
        """
        from acqdp.tensor_network.compiler import Compiler
        return Compiler(**kwargs.get('compiler_params', {})).compile(self, order)


    def refresh_signature(self):
        self.signature = hash(numpy.random.uniform())

    @property
    def nodes(self):
        return OrderedDict({n: d for n, d in self.network.nodes(data=True) if n[0] == 0})

    @property
    def edges(self):
        return OrderedDict({n: d for n, d in self.network.nodes(data=True) if n[0] == 1})

    @property
    def nodes_by_name(self):
        return [n[1] for n in self.nodes]

    @property
    def edges_by_name(self):
        return [e[1] for e in self.edges]

    @property
    def shape(self):
        """
        The common property of all :class:`TensorValued` classes, yielding the shape of the object.
        """
        return tuple([self.network.nodes[(1, e)]['dim'] for e in self.open_edges])

    def __str__(self):
        nodes_str = "Nodes: \n"
        for node in self.nodes:
            nodes_str += str(node[1]) + ": "\
                + str(self.network.nodes[node]['tensor']) + '\n'\
                + "connected to: "\
                + str([e[1] for e in self.network[node]]) + '\n'
        edges_str = "Edges: \n" + str(self.edges_by_name)
        open_str = "Open edges: " + str(self.open_edges)
        return super().__str__() + "\n"\
            + nodes_str\
            + edges_str + "\n"\
            + open_str

    @property
    def is_valid(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the :class:`TensorValued` object is valid or not.
        In every step of a program, all existing :class:`TensorValued` object must be valid, otherwise an exception should be thrown out;
        this property is for double checking that the current :class:`TensorValued` object is indeed valid.
        """
        for node in self.nodes:
            shape = [self.network.nodes[e]['dim'] for e in self.network[node]]
            lst = self.network.nodes[node]['tensor'].shape
            if lst is None:
                continue
            lst = list(lst)
            if len(lst) != len(shape):
                return False
            for i in range(len(lst)):
                if (lst[i] is not None)\
                   and (shape[i] is not None)\
                   and (lst[i] != (shape[i])):
                    return False
        return True

    @property
    def is_ready(self):
        """
        The common property of all :class:`TensorValued` classes, indicating whether the current :class:`TensorValued` object is ready for
        contraction, i.e. whether it semantically represents a tensor with a definite value.
        In the process of a program, not all :class:`TensorValued` objects need to be ready; however once the `data` property of a certain object is queried,
        such object must be ready in order to successfully yield an :class:`numpy.ndarray` object.
        """
        for node in self.nodes:
            tsr = self.network.nodes[node]['tensor']
            if not tsr.is_ready:
                return False
            shape = tuple([self.network.nodes[e]['dim'] for e in self.network[node]])
            if shape != tsr.shape:
                return False
        return True

    @property
    def closed_edges_by_name(self):
        """
        Return the set of closed edge names, i.e., edges that do not appear in the open edges.
        :returns: :class:`Set` -- The set of closed edge names.
        """
        return set(self.edges_by_name).difference(self.open_edges)

    def edges_by_name_from_nodes_by_name(self, node_names=None):
        """
        Return the list of all edges connected to any of the input nodes.

        :param node_names: a set of nodes in this tensor network.
        :type node_names: :class:`List`.

        :returns: :class:`Set` -- the set of edges who is connected to some node in the set `nodes`.
        """
        res = []
        if node_names is None:
            node_names = self.nodes_by_name
        for node_name in node_names:
            res += list(self.network[(0, node_name)])
        return set([e[1] for e in res])

    def closed_edges_by_name_from_nodes_by_name(self, node_names):
        """
        Return the list of all closed edges connected to any of the input nodes.

        :param node_names: a set of nodes in this tensor network.
        :type node_names: :class:`List`.

        :returns: :class:`Set` -- the set of closed edges who is connected to some node in the set `nodes`.
        """
        edges = self.edges_by_name_from_nodes_by_name(node_names)
        return {edge_name for edge_name in edges if set(self.network[(1, edge_name)]).issubset([(0, n) for n in node_names])
                and edge_name not in self.open_edges}

    def nodes_by_name_from_edges_by_name(self, edge_names):
        """
        Return the set of all nodes connected to the set of input edges.

        :param edge_names: a list of edges in this tensor network.
        :type edge_names: :class:`List`

        :returns: :class:`Set` -- the set of nodes who are connected to some node in the set `edges`.
        """
        res = set()
        for edge_name in edge_names:
            res |= set(self.network[(1, edge_name)])
        return set([n[1] for n in res])

    # tensor network manipulations

    def add_attribute_to_names(self, name):
        """
        Add an attribute to all the names for nodes and edges of a tensor network.

        :param name: The attribute to be added to all the names.
        :type name: :class:`Tuple` or hashable
        """
        def comb_name(att, name):
            if not isinstance(att, tuple): att = (att,)
            if not isinstance(name, tuple): name = (name,)
            return att + name
        self.network = networkx.relabel_nodes(self.network, {n: (n[0], comb_name(name, n[1])) for n in self.network.nodes}, False)
        self.open_edges = [comb_name(name, e) for e in self.open_edges]
        for node in self.nodes:
            self.network.nodes[node]['edges'] = [comb_name(name, i) for i in self.network.nodes[node]['edges']]

    def update_dimension(self, dic):
        """
        Update the dimension of the edges in the tensor network.

        :param dic: the key-value pairs of the edges to be updated. The keys are the name of the edges; the values are the corresponding updated bond dimensions. If a value is -1, that edge will be updated as a wildcard-dimensional edge, meaning that it will be automatically set a bond dimension when connected to tensor-valued objects with a definite dimension.
        :type dic: :class:`Dict`
        :raises: AssertionError -- An error will be raised when the dimension update results in inconsistent bond dimensions.
        """
        for edge_name in dic:
            if (1, edge_name) not in self.network.nodes:
                raise ValueError("Edge {} does not exist".format(edge_name))
            edge_data = self.network.nodes[(1, edge_name)]
            edge_dim = edge_data.get('dim')
            if edge_dim is None:
                edge_data['dim'] = dic[edge_name]
            else:
                if dic[edge_name] is not None:
                    assert edge_dim == dic[edge_name], "{}: {} != {}".format(edge_name,
                                                                             dic[edge_name],
                                                                             edge_dim)

    def add_node(self, node_name=None, edges=None, tensor=None, is_open=False):
        """
        Add a node into the tensor network.

        :param node_name: The name of the node to be added. If not provided, a new name will be assigned.
        :type node_name: :class:`hashable`
        :param edges: The edges the new tensor node is being connected to.
        :type edges: :class:`List` -- list of edges in the tensor network corresponding to the open edges of the new node. If not given, defaults to edges `[0, ..., len(shape)]`
        :param tensor: The value of the newly added tensor.
        :type tensor: :class:`tensor_network.tensor_valued.TensorValued`.
        :param is_open: Whether the new indices associated to the new node will be added to open_edges.
        :type is_open: :class:`bool`
        :returns: hashable -- Name of the node added. If none is given as input, return the one automatically signed.
        :raises: ValueError -- An existing node in the tensor network is being added.
        :raises: IndexError -- A node is added with ambiguous connection to the tensor network. Happens when neither the shape of the tensor nor the connecting edges are given.
        :raises: AssertionError -- A mismatch in the bond dimensions.
        """
        from .tensor import Tensor
        if (0, node_name) in self.network.nodes:
            raise ValueError("Node {} already exists in the tensor network".format(node_name))
        if not isinstance(tensor, TensorValued):
            tensor = Tensor(tensor)
        if tensor.dtype == complex:
            self.dtype = complex
        if node_name is None:
            node_name = tensor.identifier
        shape = tensor.shape
        if edges is None:
            if shape is not None:
                edges = range(len(shape))
            else:
                raise IndexError("Connection is ambiguous")
        else:
            edges = list(edges)
        self.network.add_node((0, node_name), edges=edges)
        for i, edge_name in enumerate(edges):
            bond_dim = None
            if shape is not None:
                bond_dim = shape[i]
            self.add_edge(edge_name, bond_dim)
        self.network.add_edges_from([((0, node_name), (1, edge_name)) for edge_name in edges])
        self.update_node(node_name, tensor)
        if is_open:
            for edge in edges:
                self.open_edge(edge)
        self.refresh_signature()
        return node_name

    def pop_node(self, node_name):
        """
        Pop a tensor node from the tensor network. This operation simply removes the node from the hypergraph.

        :returns: dict -- The dict containing all the information of the popped node, including its tensor and hyperedge connections.

        :param node_name: Name of the node to be popped.
        :type node_name: hashable
        """
        pop = self.network.nodes[(0, node_name)]
        self.network.remove_node((0, node_name))
        self.refresh_signature()
        return pop

    def remove_node(self, node_name):
        """
        Remove a tensor node from the tensor network. The difference between removing a node and popping anode is that popping a node does not change the
        output shape of the tensor network, while removing a node would result in a new tensor-valued object, where all edges connected to the removed node
        are appended to the end of the open edges list.

        :returns: dict -- The dict containing all the information of the removed node, including its tensor and hyperedge connections.
        :param node_name: Name of the node to be removed.
        :type node_name: hashable
        """
        pop = self.pop_node(node_name)
        self.open_edges += pop['edges']
        return pop

    def update_node(self, node_name, tensor=None):
        """
        Update an existing node of the tensor network by another tensor-valued object.

        :param node_name: Name of the node to be updated.
        :type node_name: hashable
        :param tensor: The new tensor to be put at the node. If `None`, the tensor must be updated with an actual value later in order for the tensor network to be ready for contraction.
        :type tensor: :class:`TensorValued` or `None`
        :raises: AssertionError -- Node update results in a bond dimension mismatch.
        """
        from .tensor import Tensor
        if not isinstance(tensor, TensorValued):
            tensor = Tensor(tensor)
        if tensor.dtype == complex:
            if self.dtype != complex:
                self.dtype = complex
                self.refresh_signature()
        self.network.nodes[(0, node_name)]['tensor'] = tensor
        shape = tensor.shape
        if shape is not None:
            self.update_dimension({edge: shape[i] for i, edge in enumerate(self.network.nodes[(0, node_name)]['edges'])})
        else:
            self.update_dimension({edge: None for edge in self.network.nodes[(0, node_name)]['edges']})

    def add_edge(self, edge_name=None, bond_dim=None):
        """
        Generate an index which is not in this tensor network before.

        :param edge_name: Name of the edge to be added. Will be automatically assigned if none is given.
        :type edge_name: hashable
        :param bond_dim: Bond dimension of the edge. If set to `None`, the edge bond dimension will be wildcard until set otherwise.
        :type bond_dim: `int` or `None`

        :returns: hashable -- Name of the newly added edge.
        """
        if edge_name is None:
            self.id_count += 1
            edge_name = "_" + str(self.id_count - 1)
        if (1, edge_name) not in self.network:
            self.network.add_node((1, edge_name), dim=bond_dim)
        elif bond_dim is not None:
            self.update_dimension({edge_name: bond_dim})
        return edge_name

    def open_edge(self, edge_name=None, bond_dim=None):
        """
        Append an edge to the list of open edges. If the edge does not exist, it will be created.

        :param edge_name: Name of the edge to be opened.
        :type edge_name: hashable
        :param bond_dim: Bond dimension of the edge. See `add_edge`.
        :type bond_dim: `int`
        """
        if edge_name not in self.edges_by_name:
            self.add_edge(edge_name, bond_dim)
        self.open_edges.append(edge_name)
        self.refresh_signature()

    def close_edge(self, edge_name):
        """
        Make an edge closed. All appearances of the edge will be removed from the open edges list.

        :param edge_name: Edge to be closed.
        :type edge_name: hashable
        """
        while(edge_name in self.open_edges):
            idx = self.open_edges.index(edge_name)
            self.open_edges.pop(idx)
        self.refresh_signature()

    def fix_edge(self, edge_name, fix_to=0, change_fix=False):
        """
        Fix an edge to a fixed value. The tensor network structure is not yet modified; this method only puts an attribute to the fixed hyperedges. The structure will be changed upon calling the `fix` method.

        :param edge_name: Edge to be fixed to a specific value.
        :type edge_name: hashable
        :param fix_to: the value the edge is fixed to. Should be an integer ranging from 0 to `bond_dim - 1`.
        :type fix_to: `int`
        :param change_fix: Mode of `fix_edge`. If `change_fix` is set to `True`, the edge will be fixed to the given value regardless of how it is fixed previously. If set to `False`, the fix will apply together with the previous fix, resulting in a zero tensor if the two fixes do not match.
        :type change_fix: `bool`
        """
        if isinstance(fix_to, int):
            fix_to = {fix_to}
        else:
            fix_to = set(fix_to)
        assert fix_to.issubset(range(self.network.nodes[(1, edge_name)]['dim'])), f"Invalid operation: edge being fixed to invalid value {fix_to}"
        if 'fix_to' in self.network.nodes[(1, edge_name)]:
            if not change_fix:
                self.network.nodes[(1, edge_name)]['fix_to'] = list(set(self.network.nodes[(1, edge_name)]['fix_to']) & fix_to)
            else:
                self.network.nodes[(1, edge_name)]['fix_to'][0] = list(fix_to)[0]
        else:
            self.network.nodes[(1, edge_name)]['fix_to'] = list(fix_to)
        if not change_fix:
            self.refresh_signature()

    def fix_index(self, index, fix_to=0):
        """
        Fix one outgoing index of the tensor network (see :class:`tensor_network.tensor_valued.TensorValued`).
        """
        edge = self.open_edges[index]
        self.fix_edge(edge, fix_to)
        self.open_edges.pop(index)
        self.refresh_signature()
        return self

    def fix(self):
        """
        Fix the tensor network, by removing all hyperedges attributed to a fixed value. This operation changes the structure of the tensor network without changing the value of the tensor network.
        """
        edges = list(self.edges)
        for e in edges:
            if 'fix_to' in self.network.nodes[e]:
                for node in self.nodes_by_name_from_edges_by_name([e[1]]):
                    self.network.nodes[(0, node)]['edges'] = [i for i in self.network.nodes[(0, node)]['edges'] if i != e[1]]
                if e[1] not in self.open_edges:
                    self.network.remove_node(e)
                else:
                    self.network.remove_edges_from([(e, n) for n in self.network[e]])
        return self

    def complete_delta(self):
        """
        Complete the tensor network by adding necessary delta tensors to the tensor network. This is in order to resolve the issue that repeated or new indices appear in the open edges, which is not yet recognized by Einstein summation conventions. This operation changes the structure of the tensor network without changing its value.
        """
        if '*' in self.nodes_by_name: return self
        open_edge_names = {}
        edges = self.edges_by_name_from_nodes_by_name()
        if len(self.nodes) == 0:
            id_node = numpy.array(1)
            self.add_node('*', [], id_node)
        for (i, edge) in enumerate(self.open_edges):
            if edge not in open_edge_names:
                open_edge_names.update({edge: 0})
            else:
                open_edge_names[edge] += 1
                new_node = numpy.eye(self.network.nodes[(1, edge)]['dim'])
                self.add_node(("*", edge, open_edge_names[edge]),
                            [edge, (edge, open_edge_names[edge])], new_node)
                self.open_edges[i] = (edge, open_edge_names[edge])
        for edge in open_edge_names:
            diag = None
            if edge not in edges and open_edge_names[edge] == 0:
                diag = [1] * self.network.nodes[(1, edge)]['dim']
            if 'fix_to' in self.network.nodes[(1, edge)]:
                diag = [
                    1 if i in self.network.nodes[(1, edge)]['fix_to'] else 0
                    for i in range(self.network.nodes[(1, edge)]['dim'])
                ]
                self.network.nodes[(1, edge)].pop('fix_to')
            if diag is not None:
                new_node = numpy.array(diag)
                self.add_node(('*', edge, 0), [edge], new_node)
        return self

    def merge_edges(self, edges, merge_to=None):
        """
        Merge a list of edges as one single edge by identification.

        :param edges: The list of edges to be merged into one.
        :type edges: :class:`List`
        :param merge_to: The name of the edge all the edges are merged to. If set to `None`, the merged edges will appear as `edges[0]`
        :type merge_to: hashable
        """
        dim = None
        for edge_name in edges:
            if 'dim' in self.network.nodes[(1, edge_name)]:
                dim = self.network.nodes[(1, edge_name)]['dim']
        if merge_to is None:
            merge_to = sorted(edges)[-1]
        elif (1, merge_to) not in self.network.nodes:
            self.add_edge(merge_to)
        self.update_dimension({merge_to: dim})
        nodes_list = self.nodes_by_name_from_edges_by_name(edges)
        for edge_name in edges:
            if 'fix_to' in self.network.nodes[(1, edge_name)]:
                self.fix_edge(merge_to, self.network.nodes[(1, edge_name)]['fix_to'])
            nodes = self.network[(1, edge_name)]
            self.network.add_edges_from([(node, (1, merge_to)) for node in nodes])
            if edge_name != merge_to:
                self.network.remove_node((1, edge_name))
            # replace occurrence of edge_name by merge_to
        for node in nodes_list:
            self.network.nodes[(0, node)]['edges'] =\
                [merge_to
                    if i in edges
                    else i
                    for i in self.network.nodes[(0, node)]['edges']]
        self.open_edges =\
            [merge_to
                if i in edges
                else i
                for i in self.open_edges]
        assert all((1, e) in self.network.nodes for e in self.open_edges)
        self.refresh_signature()
        return merge_to

    def rewire(self, node_name, leg, rewire_to=None):
        """
        Rewire an outgoing edge of a tensor node to another edge in the tensor network.

        :param node_name: Name of the tensor node for whom an outgoing edge is to be rewired.
        :type node_name: hashable
        :param leg: Index of the outgoing edge regarding the tensor node. Should be an integer ranging from `0` to rank(t) - 1, t being the corresponding tensor.
        :type leg: `int`
        :param rewire_to: The edge name in the tensor network to redirect the leg to. If `None`, it will be rewired to a newly created edge.
        :type rewire_to: hashable
        """
        dim = None
        if self.network.nodes[(0, node_name)]['tensor'].shape is not None:
            dim = self.network.nodes[(0, node_name)]['tensor'].shape[leg]
        edge_name = self.network.nodes[(0, node_name)]['edges'][leg]
        if rewire_to not in self.edges_by_name:
            rewire_to = self.add_edge(rewire_to, dim)
        self.network.nodes[(0, node_name)]['edges'][leg] = rewire_to
        self.network.add_edge((0, node_name), (1, rewire_to))
        if edge_name not in self.network.nodes[(0, node_name)]['edges']:
            self.network.remove_edge((0, node_name), (1, edge_name))
        self.refresh_signature()

    def expand(self, nodes=None, recursive=False):
        """
        Expand all tensor network hierarchical structure in the tensor network. Specifically, when there is a tensor node in the tensor network which is a
        :class:`TensorNetwork` or a :class:`TensorView` object, the tensor nodes in the tensor node are lifted up to the current tensor network level with
        corresponding connections.
        """
        from .tensor_view import TensorView
        if nodes is None:
            def cm(a, b):
                try:
                    return a < b
                except TypeError:
                    try:
                        return a < (b,)
                    except TypeError:
                        try:
                            return (a,) < b
                        except TypeError:
                            return str(a) < str(b)
            nodes = sorted(list(self.nodes_by_name), key=cmp_to_key(cm))
        for node_name in nodes:
            if isinstance(self.network.nodes[(0, node_name)]['tensor'], TensorView):
                self.update_node(node_name, self.nodes_by_name[node_name]['tensor'].expand(recursive))
            if issubclass(type(self.network.nodes[(0, node_name)]['tensor']), TensorNetwork):
                pop = self.pop_node(node_name)
                tsr = pop['tensor'].copy()
                tsr.expand(recursive=True)
                connection = pop['edges']
                tsr.add_attribute_to_names(node_name)
                self.network.add_nodes_from(tsr.network.nodes(data=True))
                self.network.add_edges_from(tsr.network.edges)
                open_edges = tsr.open_edges
                g = networkx.Graph()
                g.add_edges_from([(connection[i], open_edges[i]) for i in range(len(open_edges))])
                for ls in networkx.connected_components(g):
                    merge_to = sorted(list(set(ls).intersection(connection)))[-1]
                    self.merge_edges(list(ls), merge_to)
        self.refresh_signature()
        return self

    def subtn(self, nodes=None, edges=None, open_edges=None):
        if nodes is None:
            if edges is None:
                return self.nodes_by_name, self.closed_edges_by_name, self.copy()
            else:
                nodes = self.nodes_by_name_from_edges_by_name(edges)
        else:
            assert all((0, n) in self.network.nodes for n in nodes), f"Illegal sub tensor network {nodes, [(0, n) in self.network.nodes for n in nodes], self.network.nodes}"
            if edges is None:
                edges = self.closed_edges_by_name_from_nodes_by_name(nodes)
            else:
                assert edges.issubset(self.closed_edges_by_name_from_nodes_by_name(nodes)),\
                    "Illegal sub tensor network"
        all_open_edges = [e for e in self.edges_by_name_from_nodes_by_name(nodes).difference(edges) if 'fix_to' not in self.network.nodes[(1, e)]]
        if open_edges is None:
            open_edges = all_open_edges
        else:
            assert(set(open_edges) == set(all_open_edges)), "Illegal sub tensor network"

        stn = TensorNetwork(open_edges=open_edges, dtype=self.dtype)
        for node_name in nodes:
            stn.add_node(node_name, self.network.nodes[(0, node_name)]['edges'], self.network.nodes[(0, node_name)]['tensor'])
        for edge in edges:
            for e in self.network.nodes[(1, edge)]:
                stn.network.nodes[(1, edge)][e] = self.network.nodes[(1, edge)][e]
        return nodes, edges, stn

    def encapsulate(self,
                    nodes=None,
                    edges=None,
                    open_edges=None,
                    stn_name=None):
        nodes, edges, stn = self.subtn(nodes, edges, open_edges)
        for node_name in nodes:
            self.pop_node(node_name)
        self.network.remove_nodes_from([(1, edge_name) for edge_name in edges])
        stn_name = self.add_node(stn_name, stn.open_edges, stn)
        self.refresh_signature()
        return stn_name, stn

    def partial_contract(self,
                         nodes=None,
                         edges=None,
                         open_edges=None,
                         stn_name=None):
        stn_name, stn = self.encapsulate(nodes, edges, open_edges, stn_name)
        self.update_node(stn_name, stn.contract())

    def simplify(self, recursive=True):
        flag = True
        while flag:
            flag = False
            for edge in self.edges:
                found = False
                for n0, n1 in itertools.combinations(self.network[edge], 2):
                    fix = len([e for e in set(self.network[n0]).intersection(set(self.network[n1])) if len(self.network[e]) == 2])
                    sz = max(len(self.network[n0]), len(self.network[n1]))
                    new_sz = len(set(self.network[n0]).union(self.network[n1])) - fix
                    if new_sz <= sz:
                        flag = recursive
                        self.encapsulate(nodes=[n0[1], n1[1]])
                        found = True
                        break
                if found:
                    break

    def _expand_and_delta(self):
        return self.copy().expand(recursive=True).complete_delta()

    @property
    def tanner_graph(self):
        nodes = list(self.nodes)
        edges = [e for e in self.open_edges if 'fix_to' not in self.network.nodes[(1,e)]]
        edges += [e for e in self.closed_edges_by_name if 'fix_to' not in self.network.nodes[(1,e)]]
        x_nodes = []
        y_nodes = []
        for i, node in enumerate(nodes):
            edge_list  = [i for i in self.network.nodes[node]['edges'] if i in edges]
            x_nodes += [i+1] * len(edge_list)
            for edge in edge_list:
                y_nodes.append(edges.index(edge))
        edge_list = [j for j,i in enumerate(edges) if i in self.open_edges]
        y_nodes += edge_list
        x_nodes += [0] * len(edge_list)
        res = sparse.csr_matrix(([True] * len(x_nodes), (x_nodes, y_nodes)), dtype=bool)
        return res, [n[1] for n in nodes]

    def subscripts(self, nodes=None):
        if nodes is None:
            nodes = list(self.nodes_by_name)
        edges = [e for e in self.edges_by_name if 'fix_to' not in self.network.nodes[(1, e)]]
        lhs = []
        shapes = []
        for node in nodes:
            lhs.append(''.join([opt_einsum.get_symbol(edges.index(e)) for e in self.network.nodes[(0, node)]['edges'] if e in edges]))
            shapes.append(list(2 if self.network.nodes[(1, e)]['dim'] is None else self.network.nodes[(1, e)]['dim'] for e in self.network.nodes[(0, node)]['edges'] if e in edges))
        rhs = ''.join([opt_einsum.get_symbol(edges.index(e)) for e in self.open_edges if e in edges])
        return lhs, rhs, shapes

    def copy(self):
        tn = TensorNetwork(copy.copy(self.open_edges), dtype=self.dtype)
        tn.network = self.network.copy()
        for node in tn.nodes:
            tn.network.nodes[node]['tensor'] = tn.network.nodes[node]['tensor']
        tn.signature = self.signature
        return tn

    def __deepcopy__(self, memo):
        tn = TensorNetwork(copy.deepcopy(self.open_edges), dtype=self.dtype)
        tn.network = copy.deepcopy(self.network)
        for node in tn.nodes:
            tn.update_node(node[1], copy.deepcopy(tn.network.nodes[node]['tensor']))
        tn.signature = self.signature
        return tn

    def cast(self, dtype):
        for node in self.nodes:
            self.network.nodes[node]['tensor'] = self.network.nodes[node]['tensor'].cast(dtype)
        return self



import kahypar
import opt_einsum as oe
import numpy
from numpy import random
import itertools
import copy
from os.path import join, abspath, dirname
import cma
import os
import tempfile
import pickle
from multiprocessing import Pool
import cProfile
from scipy import sparse
from acqdp.utility import opt_einsum_ext

from acqdp.tensor_network.local_optimizer import OrderResolver
from acqdp.tensor_network.contraction import ContractionScheme, ContractionCost, OrderCounter
from acqdp.tensor_network.order_finder import OrderFinder

CMA_TEMP_DIR = tempfile.TemporaryDirectory()
CMA_FILE_PREFIX = CMA_TEMP_DIR.name + os.sep

KAHYPAR_PROFILE_DIR = join(abspath(dirname(__file__)), 'kahypar_profiles')

modes = ['direct', 'recursive']
objectives = ['cut', 'km1']
queryOrderResolver = OrderResolver(optimal=9, dp=9)


def query(order_finder,
          s,
          eps,
          graph_copy,
          seed=numpy.random.randint(2**31 - 1),
          counter=None,
          objective=0,
          cutoff=25):
    K = s // 2 + 3
    mode = s % 2
    tg, nodes = graph_copy.tanner_graph
    order = order_finder._order_by_params(numpy.array(tg.todense()),
                                         copy.copy(nodes),
                                         K=K,
                                         eps=eps,
                                         mode=mode,
                                         objective=0,
                                         cutoff=cutoff,
                                         seed=seed,
                                         counter=counter)
    res = queryOrderResolver.order_to_contraction_scheme(graph_copy, order)
    return res


class KHPOrderFinder(OrderFinder):
    """
    :class:`KHPOrderFinder` utilizes hypergraph decomposition and cma-es opmization scheme for finding contraction orders for intermediate-size tensor networks.

    :ivar num_threads: Number of threads to concurrently search for the best order.
    :ivar num_iters: Number of iterations for cma optimization.
    :ivar num_cmas: Number of times cma-es optimization scheme is called for exploring better parameter combinations.
    :ivar cma_args: Arguments for cma-es optimization scheme. See the documentation of cma for more details
    :ivar eps: In case a good parameter combination is already known, inputting a eps would skip the cma optimization scheme and directly yield orders found by hypergraph decomposition.
    """

    def __init__(
            self,
            num_threads=28,
            num_iters=50,
            num_cmas=1,
            cma_args={
                'bounds': [0.5, 0.5, 0.5],
                'std_dev': 0.2,
                'kwargs': {
                    'seed': 1175,
                    'bounds': [[0.01, 0, 0], [1, 1, 1]],
                    'tolfun': 1e-2,
                    'verb_filenameprefix': CMA_FILE_PREFIX
                },
            },
            eps=None,
            **kwargs):
        self.num_threads = num_threads
        self.num_iters = num_iters
        self.num_cmas = num_cmas
        self.eps = eps
        self.cma_args = cma_args

    def _simplify(self, graph):
        order = []
        counter = OrderCounter()
        flag = True
        while flag:
            flag = False
            for edge in graph.edges:
                found = False
                for n0, n1 in itertools.combinations(graph.network[edge], 2):
                    s0, s1 = set(graph.network[n0]), set(graph.network[n1])
                    fix = len([
                        e for e in s0.intersection(s1)
                        if len(graph.network[e]) == 2
                    ])
                    sz = max(len(s0), len(s1))
                    new_sz = len(s0.union(s1)) - fix
                    if new_sz <= sz:
                        flag = True
                        new_name = counter.cnt
                        new_name, _ = graph.encapsulate(nodes=[n0[1], n1[1]],
                                                        stn_name=new_name)
                        order.append([[n0[1], n1[1]], new_name])
                        found = True
                        break
                if found:
                    break
        return order, counter

    def find_order(self, graph):
        graph_copy = graph._expand_and_delta()
        graph_copy.fix()
        init_order, counter = self._simplify(graph_copy)
        if self.eps is None:
            seed_f = numpy.random.uniform(0, 1)

            def probe(x):
                eps = [x[0], 0.95 + 0.049 * x[1]]
                with Pool(self.num_threads) as p:
                    reses = p.starmap(
                        query,
                        [(self, s, eps, graph_copy, int(
                            (2**31 - 1) * seed), copy.copy(counter))
                         for s in range(self.num_threads)
                         for seed in x[2:]])
                res = min(reses, key=lambda x: (x.cost, reses.index(x)))
                return res

            curr_y = None
            for _ in range(self.num_cmas):
                seed_f = numpy.random.uniform(0, 1)
                self.cma_args['kwargs']['seed'] = numpy.random.randint(2**31 -
                                                                       1)

                def probe_value(x):
                    res = probe([x[0], x[1], seed_f])
                    return numpy.log10(float(res.cost.s))

                es = cma.CMAEvolutionStrategy(
                    self.cma_args['bounds'], self.cma_args['std_dev'],
                    self.cma_args['kwargs']).optimize(probe_value,
                                                      iterations=self.num_iters)
                x = es.result[0]
                y = es.result[1]
                if curr_y is None or curr_y > y:
                    self.eps = [x[0], 0.95 + 0.049 * x[1]]
                    curr_y = y
            res = probe(x)
            yield ContractionScheme(init_order + res.order, cost=res.cost)
        while True:
            reses = [
                query(self,
                      s,
                      self.eps,
                      graph_copy,
                      seed=numpy.random.randint(2**31 - 1),
                      counter=copy.copy(counter)) for s in range(10)
            ]
            res = min(reses, key=lambda x: (x.cost, reses.index(x)))
            yield ContractionScheme(init_order + res.order, cost=res.cost)

    def _order_by_params(self, graph, nodes, **kwargs):
        graph = self._refresh(graph)
        counter = kwargs.get('counter')

        K = int(kwargs.get('K'))
        eps = kwargs.get('eps', None)
        top = kwargs.get('top', True)
        kwargs_c = copy.copy(kwargs)
        kwargs_c['eps'] = eps[0]
        if not top:
            kwargs_c['K'] = 2
            kwargs_c['eps'] = eps[1]
        context = self._set_context(**kwargs_c)
        cutoff = kwargs.get('cutoff')
        hypergraph, _, edges = self._init_hypergraph_tanner(graph, kwargs_c['K'])
        if not edges:
            return [[nodes, '#']]
        kahypar.partition(hypergraph, context)
        partitions_names = [[] for _ in range(K)]
        for i, n in list(enumerate(nodes)):
            partitions_names[hypergraph.blockID(i)].append(n)
        if any(len(part) == len(nodes) for part in partitions_names):
            return [[nodes, '#']]
        order = []
        all_names = []
        for (i, partite_nodes) in enumerate(
                sorted(partitions_names, key=lambda x: -len(x))):
            if len(partite_nodes) == 0:
                continue
            elif len(partite_nodes) > 1:
                new_stn = counter.cnt
                all_names.append(new_stn)
                new_g, graph, nodes = self._sub(graph, nodes, partite_nodes,
                                               new_stn)
                assert graph.shape[0] == len(nodes) + 1, (graph.shape[0],
                                                          len(nodes))
                kwargs['top'] = False
                if len(partite_nodes) > cutoff:
                    new_order = self._order_by_params(new_g, partite_nodes,
                                                     **kwargs)
                    new_order[-1][1] = new_stn
                else:
                    new_order = [[partite_nodes, new_stn]]
                order += new_order
            else:
                all_names.append(partite_nodes[0])
        order.append([all_names, '#'])
        return order

    def _set_context(self, **kwargs):
        mode = modes[int(kwargs.get('mode'))]
        objective = objectives[int(kwargs.get('objective'))]
        K = int(kwargs.get('K'))
        eps = kwargs.get('eps')
        seed = kwargs.get('seed')
        profile_mode = {'direct': 'k', 'recursive': 'r'}[mode]
        profile = f"{objective}_{profile_mode}KaHyPar_sea20.ini"
        context = kahypar.Context()
        context.loadINIconfiguration(join(KAHYPAR_PROFILE_DIR, profile))
        context.setK(K)
        context.setSeed(seed)
        context.setEpsilon(kwargs.get('epsilon', eps * (K - 1)))
        context.suppressOutput(kwargs.get('quiet', True))
        return context

    def _sub(self, graph, nodes, sub_nodes, new_name):
        inds = [False] + [n in sub_nodes for n in nodes]
        out_nodes = [n for n in nodes if n not in sub_nodes] + [new_name]
        subb = graph[inds]
        rest = graph[numpy.invert(inds)]
        open_edges = numpy.any(subb, axis=0) & numpy.any(rest, axis=0)
        return numpy.append(open_edges[None, :], subb,
                            axis=0), numpy.append(rest,
                                                  open_edges[None, :],
                                                  axis=0), out_nodes

    def _init_hypergraph_tanner(self, graph, k=2):
        nodes = list(range(1, graph.shape[0]))
        edges = list(range(graph.shape[1]))
        hyperedge_indices = [0]
        hyperedges = []
        edge_weights = []
        for edge in range(graph.shape[1]):
            hyperedges += list(numpy.where(graph[1:, edge])[0])
            hyperedge_indices.append(len(hyperedges))
            edge_weights.append(1)
        hp = kahypar.Hypergraph(len(nodes), len(edges), hyperedge_indices,
                                hyperedges, k, edge_weights, []), nodes, edges
        return hp

    def _refresh(self, graph):
        return graph[:, numpy.sum(graph, axis=0) > 0]



from acqdp.circuit.circuit import Channel, Circuit
import numpy as np

def Depolarization(px=0.25, py=None, pz=None):
    """
    Single-qubit depolarizing channel.

    :param px: One of the three parameters to describe the single-qubit depolarizing channel.
    :type px: :class:`float`.
    :param py: One of the three parameters to describe the single-qubit depolarizing channel.
    :type py: :class:`float`.
    :param pz: One of the three parameters to describe the single-qubit depolarizing channel.
    :type pz: :class:`float`.
    :returns: :class:`Mixture` -- the single-qubit depolarizing channel.
    """
    if py is None:
        py = px
    if pz is None:
        pz = px
    res = Channel(
        1,
        np.array([[[[1 - px - py, 0], [0, 1 - 2 * pz - px - py]],
                   [[0, px + py], [px - py, 0]]],
                  [[[0, px - py], [px + py, 0]],
                   [[1 - px - py - 2 * pz, 0], [0, 1 - px - py]]]]))
    return res


def Dephasing(pz=0.5):
    """
    Single-qubit dephasing channel, also called the phase-damping channel.

    :param pz: The parameter to describe the single-qubit dephasing channel.
    :type pz: :class:`float`.
    :returns: :class:`Mixture` -- the single-qubit dephasing channel with given parameter.
    """
    return Depolarization(0, 0, pz)


def AmplitudeDampling(p=0.1):
    res = Channel(
        1,
        np.array([[[[1, 0], [0, np.sqrt(1-p)]],
                   [[0, p], [0, 0]]],
                  [[[0, 0], [0, 0]],
                   [[np.sqrt(1-p), 0], [0, 1-p]]]]))
    return res


def add_noise(circuit, noise_channel):
    c = Circuit()
    for k in circuit.operations_by_name:
        operation = circuit.operations_by_name[k]['operation']
        qubits = circuit.operations_by_name[k]['qubits']
        time_step = circuit.operations_by_name[k]['time_step']
        new_op = Circuit() | operation
        for a in new_op._output_indices[0]:
            new_op.append(noise_channel, [a])
        c.append(new_op, qubits=qubits, time_step=time_step, name=k)
    return c
